"""Module containing the Profiles class for storing measurements.

:Author(s): **Christoffer Hjertø Grindheim**"""
from __future__ import annotations
import logging

import typing as t
import collections.abc as col
import numpy as np
import scipy.signal as sig
from scipy import constants

from .. import exceptions as expt
from ..utils import physics

log = logging.getLogger(__name__)


class Profiles:
    """Class holding measured data.

    This class holds the measured data (waterfall) and their properties.

    The waterfall is a nprofiles x nbins shaped array, waterfall[0] is the
    first measurement. Each bin is `Machine.dtbin` long,
    and holds the intensity of the beam at this time of the measurement.

    The profile charge can be provided, or calculated using
    the `Profiles.calc_profilecharge` function which is based
    on the machines pickup sensitivity.

    Also, the class can be used to calculate the self-fields of the bunch
    by calling the `Profiles.calc_self_fields` function.

    Parameters
    ----------
    machine: Machine
        Holds information about the measurements and the machine.
    sampling_time: float
        Original measurement sampling time.
    waterfall: ndarray
        2D-array containing all profile measurements.
    profile_charge: float, optional, default=None
        Total charge of a reference profile.

    Attributes
    ----------
    machine: Machine
        The machine and its settings when measurements was taken.
        This object is needed for assertions, to check that the provided
        waterfall is correct compared to the machine parameters. It is
        also needed for calculating the self fields and profile charge
        of the bunch.
    sampling_time: float
        Original sampling time of measurements.
        Needed for calculation of profile charge.
    waterfall: ndarray
        2D array containing all profile measurements.
        Array should have the shape: (N, M), where N is the number of profiles
        and M is the number of bins of each profile.
    profile_charge: float
        The total charge of a reference profile.
    vself: ndarray, float
        2D array of self-fields at each bin of each profile.
        The shape of the array is (N, W), N is the number of profiles, and
        W is the `Profiles.wrap_length`.
    dsprofiles: ndarray, float
        2D array containing Filtered profiles. Shape should be
        (N, M), where N is the number of profiles
        and M is the number of bins of each profile.
    phiwrap: float
        The phase [rad] covering an integer of rf periods.
    wrap_length: float
        Maximum number of bins to cover an integer number of rf periods.
    """

    def __init__(self, machine, sampling_time, waterfall,
                 profile_charge=None):
        self.machine = machine
        self.sampling_time = sampling_time

        self.waterfall = waterfall

        self.profile_charge = profile_charge

        # init
        self.dsprofiles = None
        self.phiwrap = None
        self.wrap_length = None
        self.vself = None

    @property
    def waterfall(self) -> np.ndarray:
        """Waterfall defined as @property. The property does the
        following when accessed:

        * Asserts for correctness of waterfall, here the number of profiles\
        should be as stated by the `Machine` object.
        * Negative values of waterfall are set to zero.
        * `Machine.nbins` is updated with the number of bins in the waterfall.

        Parameters
        ----------
        waterfall: ndarray
            Measured profiles as a 2D array. Shape: (N, M), N is the number
            of profiles and M is the number of bins of each profile.

        Returns
        -------
        waterfall: ndarray
            Measured profiles as a 2D array. Shape: (N, M), N is the number
            of profiles and M is the number of bins of each profile.

        Raises
        ------
        WaterfallError: Exception
            Waterfall not iterable or wrong has the amount of profiles.
        WaterfallReducedToZero: Exception
            All of waterfall is reduced to zero after removal of negative
            values.
        """
        return self._waterfall

    @waterfall.setter
    def waterfall(self, new_waterfall):
        log.info('Loading waterfall...')

        if not hasattr(new_waterfall, '__iter__'):
            raise expt.WaterfallError('waterfall should be an iterable.')

        if not new_waterfall.shape[0] == self.machine.nprofiles:
            err_msg = f'Waterfall does not correspond to machine object.\n' \
                      f'Expected nr of profiles: {self.machine.nprofiles}, ' \
                      f'nr of profiles in waterfall: {new_waterfall.shape[0]}'
            raise expt.WaterfallError(err_msg)

        new_waterfall = np.array(new_waterfall)
        self._waterfall = new_waterfall.clip(0.0)
        if np.sum(np.abs(self.waterfall)) == 0.0:
            raise expt.WaterfallReducedToZero()

        self.machine.nbins = self._waterfall.shape[1]
        log.info(f'Waterfall loaded with shape: {self.waterfall.shape})')

    def calc_profilecharge(self):
        """Calculate the total charge of profile.
        Uses the beam reference profile to decide which profile should
        be used for the calculation.
        Sets the `Profiles.profile_charge` field.

        **NB:** The function requires that the **original sampling time**
        of the measurements is provided.
        """
        ref_prof = self.waterfall[self.machine.beam_ref_frame]
        self.profile_charge = (np.sum(ref_prof) * self.sampling_time
                               / (constants.e
                                  * self.machine.pickup_sensitivity))

    def calc_self_fields(self, filtered_profiles: np.ndarray = None,
                         in_filter: t.Callable = None):
        """Calculate self-fields based on filtered profiles.
        If filtered profiles are not provided by the user,
        standard filter (savitzky-golay smoothing filter) is used.

        The function sets the following fields:

        * vself - 2D array of self-fields at each bin of each profile.
        * dsprofiles - Filtered profiles.
        * phiwrap - Phase covering an integer of rf periods.
        * wrap_length - Maximum number of bins to cover\
                        an integer number of rf periods.

        A description of these attributes can be found in the documentation\
        for the `Profiles` class.

        Parameters
        ----------
        filtered_profiles: ndarray (optional, default=None)
            If the filtered profiles are provided, they will be used in the\
            calculation of the self fields. If not, the original profiles\
            will be filtered using a standard or user spescified filter.
        in_filter: function (optional, default=None)
            If provided, the measured profiles will be filtered using
            the provided filter in stead of the savitzky-golay
            smoothing filter.

        Raises
        ------
        ProfileChargeNotCalculated: Exception
            No profile charge has been provided or calculated.
        FilteredProfilesError: Exception
            Filtered profiles has the wrong shape or are not iterable.
        """
        if self.profile_charge is None:
            err_msg = 'Profile charge must be calculated before ' \
                      'calculating the self-fields'
            raise expt.ProfileChargeNotCalculated(err_msg)

        if filtered_profiles is not None:
            self.dsprofiles = self._check_manual_filtered_profs(
                filtered_profiles)
        elif in_filter is not None:
            self.dsprofiles = np.copy(self.waterfall)
            self.dsprofiles = in_filter(self.dsprofiles)
            self.dsprofiles = self._check_manual_filtered_profs(
                self.dsprofiles)
        else:
            self.dsprofiles = np.copy(self.waterfall)
            # Makes normalized version of waterfall
            self.dsprofiles /= np.sum(self.dsprofiles, axis=1)[:, None]
            self.dsprofiles = sig.savgol_filter(
                x=self.dsprofiles, window_length=7,
                polyorder=4, deriv=1)

        (self.phiwrap,
         self.wrap_length) = self._find_wrap_length()

        self.vself = self._calculate_self()
        log.info('Self fields were calculated.')

    # Checks the filtered profiles
    def _check_manual_filtered_profs(self, fprofs: col.Sequence) -> np.ndarray:
        if not hasattr(fprofs, '__iter__'):
            err_msg = 'Filtered profiles should be iterable.'
            raise expt.FilteredProfilesError(err_msg)
        fprofs = np.array(fprofs)
        if fprofs.shape == self.waterfall.shape:
            return fprofs
        else:
            err_msg = f'Filtered profiles should have the same shape' \
                      f'as the waterfall.\n' \
                      f'Shape profiles: {fprofs.shape}\n' \
                      f'Shape waterfall: {self.waterfall.shape}\n'
            raise expt.FilteredProfilesError(err_msg)

    # Calculate the number of bins in the first
    # integer number of rf periods, larger than the image width.
    def _find_wrap_length(self) -> t.Tuple[float, int]:
        if self.machine.bdot > 0.0:
            last_turn_index = ((self.machine.nprofiles - 1)
                               * self.machine.dturns - 1)
            drad_bin = (self.machine.h_num
                        * self.machine.omega_rev0[last_turn_index]
                        * self.machine.dtbin)
        else:
            drad_bin = (self.machine.h_num
                        * self.machine.omega_rev0[0]
                        * self.machine.dtbin)

        phiwrap = np.ceil(self.machine.nbins * drad_bin
                          / (2 * np.pi)) * 2 * np.pi

        wrap_length = int(np.ceil(phiwrap / drad_bin))

        log.info(f'phi wrap =  {str(phiwrap)}, '
                 f'wrap length =  {str(wrap_length)}')
        return phiwrap, wrap_length

    # Calculate self-field voltages
    def _calculate_self(self) -> np.ndarray:
        sfc = physics.calc_self_field_coeffs(self.machine)
        vself = np.zeros((self.machine.nprofiles - 1,
                          self.wrap_length + 1),
                         dtype=float)
        for i in range(self.machine.nprofiles - 1):
            vself[i, :self.machine.nbins] = (0.5 * self.profile_charge
                                             * (sfc[i]
                                                * self.dsprofiles[
                                                  i, :self.machine.nbins]
                                                + sfc[i + 1]
                                                * self.dsprofiles[
                                                  i + 1, :self.machine.nbins]))
        return vself



"""Module containing the TomographyCpp class

:Author(s): **Christoffer Hjertø Grindheim**, **Anton Lu**
"""
from __future__ import annotations
import logging
import time as tm
import typing as t
import sys

import numpy as np
import matplotlib.pyplot as plt

from .__tomography import TomographyABC
from ..cpp_routines import libtomo
from .. import exceptions as expt
from longitudinal_tomography.utils import tomo_config as conf

log = logging.getLogger(__name__)


class Tomography(TomographyABC):
    """Class for performing tomographic reconstruction of phase space.

    The tomographic routine largely consists of two parts. Projection and
    back projection. The **back projection** creates a phase space
    reconstruction based on the measured profiles. The **projection**
    routine converts from back projection to reconstructed profiles.

    By comparing the reconstructed profiles to the measured profiles,
    adjustments of the weights can be made in order to create
    a better reconstruction. The number of iterations in this process can be
    specified by the user.

    Parameters
    ----------
    waterfall: ndarray
        2D array of measured profiles, shaped: (nprofiles, nbins).
    x_coords: ndarray
        x-coordinates of particles, given as coordinates of the reconstructed
        phase space coordinate system. Shape: (nparts, nprofiles).

    Attributes
    ----------
    nparts: int
        Number of test particles.
    nprofs: int
        Number of profiles (time frames).
    nbins: int
        Number of bins in each profile.
    waterfall: ndarray
        2D array of measured profiles, shaped: (nprofiles, nbins).
        Negative values of waterfall is set to zero, and the waterfall is
        normalized.
    xp: ndarray
        x-coordinates of particles, given as coordinates of the reconstructed
        phase space coordinate system. Shape: (nparts, nprofiles).
    recreated: ndarray
        Recreated waterfall. Directly comparable with *Tomography.waterfall*.
        Shape: (nprofiles, nbins).
    diff: ndarray
        Discrepancy for phase space reconstruction at each iteration
        of the reconstruction process.
    """

    def __init__(self, waterfall: np.ndarray, x_coords: np.ndarray = None,
                 y_coords: np.ndarray = None):
        super().__init__(waterfall, x_coords, y_coords)

    def run_hybrid(self, niter=20, verbose=False):
        """Function to perform tomographic reconstruction, implemented
        as a hybrid between C++ and Python.

        Projection and back projection routines are called from C++,
        the rest is written in python. Kept for reference.

        The discrepancy of each iteration is saved in the *diff* array
        of the object.

        The recreated waterfall can be found calling the *recreated* field
        of the object.

        Parameters
        ----------
        niter: int
            Number of iterations in reconstruction.
        verbose: boolean
            Flag to indicate that the status of the tomography should be
            written to stdout. The output is identical to output
            generated in the original Fortran tomography.

        Returns
        -------
        weight: ndarray
            1D array containing the weight of each particle.

        Raises
        ------
        CoordinateError: Exception
            X-coordinates is None
        WaterfallReducedToZero: Exception
            All of reconstructed waterfall reduced to zero.
        """
        log.warning('TomographyCpp.run_hybrid() '
                    'may be removed in future updates!')
        if self.xp is None:
            raise expt.CoordinateError(
                'x-coordinates has value None, and must be provided')

        self.diff = np.zeros(niter + 1)
        reciprocal_pts = self._reciprocal_particles()
        flat_points = self._create_flat_points()
        flat_profs = np.ascontiguousarray(
            self.waterfall.flatten()).astype(np.float64)
        weight = np.zeros(self.nparts)

        weight = libtomo.back_project(weight, flat_points, flat_profs,
                                      self.nparts, self.nprofs)
        weight = weight.clip(0.0)

        if verbose:
            print(' Iterating...')

        for i in range(niter):
            if verbose:
                print(f'{i + 1:3d}')

            self.recreated = self._project(flat_points, weight)

            diff_waterfall = self.waterfall - self.recreated
            self.diff[i] = self._discrepancy(diff_waterfall)

            # Weighting difference waterfall relative to number of particles
            diff_waterfall *= reciprocal_pts.T

            weight = libtomo.back_project(
                weight, flat_points, diff_waterfall.flatten(),
                self.nparts, self.nprofs)
            weight = weight.clip(0.0)

        self.recreated = self._project(flat_points, weight)

        # Calculating final discrepancy
        diff_waterfall = self.waterfall - self.recreated
        self.diff[-1] = self._discrepancy(diff_waterfall)

        if verbose:
            print(' Done!')

        return weight
    
    def run_hybrid_multi(self, centers, cuts, niter=20, verbose=False):
        """Function to perform tomographic reconstruction, implemented
        as a hybrid between C++ and Python.

        Projection and back projection routines are called from C++,
        the rest is written in python. Kept for reference.

        The discrepancy of each iteration is saved in the *diff* array
        of the object.

        The recreated waterfall can be found calling the *recreated* field
        of the object.

        Parameters
        ----------
        niter: int
            Number of iterations in reconstruction.
        verbose: boolean
            Flag to indicate that the status of the tomography should be
            written to stdout. The output is identical to output
            generated in the original Fortran tomography.

        Returns
        -------
        weight: ndarray
            1D array containing the weight of each particle.

        Raises
        ------
        CoordinateError: Exception
            X-coordinates is None
        WaterfallReducedToZero: Exception
            All of reconstructed waterfall reduced to zero.
        """
        log.warning('TomographyCpp.run_hybrid() '
                    'may be removed in future updates!')
        if self.xp is None:
            raise expt.CoordinateError(
                'x-coordinates has value None, and must be provided')

        nBunches = len(centers)

        masks = []
        for cent, cut in zip(centers, cuts):
            mask = np.all(((self.xp+cent)>cut[0])*((self.xp+cent)<cut[1]),
                          axis=1)
            masks.append(mask)
        masks = np.array(masks)

        print(1)
        t0 = tm.process_time()

        self.diff = np.zeros(niter + 1)
        reciprocal_pts = self._reciprocal_particles_multi(centers)
        print(f"recip: {tm.process_time() - t0}")
        t0 = tm.process_time()
        flat_points = self._create_flat_points()
        print(f"flat_points: {tm.process_time() - t0}")

        # print(reciprocal_pts.shape)
        # plt.plot(reciprocal_pts[-1])
        # plt.gca().twinx().plot(self.waterfall[-1])
        # plt.xlim([0, 20])
        plt.show()


        print(2)

        flat_profs = np.ascontiguousarray(
            self.waterfall.flatten()).astype(np.float64)
        weight = np.zeros([nBunches, self.nparts])

        print(3)
        t0 = tm.process_time()
        for i in range(nBunches):
            print(f"weight {i}")
            weight[i][masks[i]] = libtomo.back_project(weight[i][masks[i]],
                                                       flat_points[masks[i]]\
                                                          + centers[i],
                                                       flat_profs,
                                                       np.sum(masks[i]),
                                                       self.nprofs)
        weight = weight.clip(0.0)
        print(f"weights 1: {tm.process_time() - t0}")
        if verbose:
            print(' Iterating...')

        print("start iter")

        for i in range(niter):
            if verbose:
                print(f'{i + 1:3d}')

            self.full_recreated = np.zeros_like(self.waterfall)

            for j in range(nBunches):
                self.recreated = self._project_multi(flat_points + centers[j],
                                                      weight[j], self.nparts)

                self.full_recreated += self.recreated

            self.recreated = self._normalize_profiles(self.full_recreated)

            diff_waterfall = self.waterfall - self.recreated
            self.diff[i] = self._discrepancy(diff_waterfall)

            # Weighting difference waterfall relative to number of particles
            diff_waterfall *= reciprocal_pts

            for j in range(nBunches):

                weight[j][masks[j]] = libtomo.back_project(
                                            weight[j][masks[j]],
                                            flat_points[masks[j]] + centers[j],
                                            diff_waterfall.flatten(),
                                            np.sum(masks[j]), self.nprofs)

            weight = weight.clip(0.0)

        self.full_recreated = np.zeros_like(self.waterfall)
        
        for j in range(nBunches):
            self.recreated = self._project_multi(flat_points + centers[j],
                                                  weight[j], self.nparts)
            self.full_recreated += self.recreated
        self.recreated = self._normalize_profiles(self.full_recreated)

        # Calculating final discrepancy
        diff_waterfall = self.waterfall - self.recreated
        self.diff[-1] = self._discrepancy(diff_waterfall)

        if verbose:
            print(' Done!')

        return weight


    # Project using C++ routine from tomolib_wrappers.
    # Normalizes recreated profiles before returning them.
    def _project(self, flat_points: np.ndarray, weight: np.ndarray) \
            -> np.ndarray:
        # rec = tlw.project(np.zeros(self.recreated.shape), flat_points,
        #                   weight, self.nparts, self.nprofs, self.nbins)
        rec = libtomo.project(np.zeros(self.recreated.shape), flat_points,
                              weight, self.nparts, self.nprofs, self.nbins)
        rec = self._normalize_profiles(rec)
        return rec

    def _project_multi(self, flat_points: np.ndarray, weight: np.ndarray,
                       nUseParts: int) -> np.ndarray:
        # rec = tlw.project(np.zeros(self.recreated.shape), flat_points,
        #                   weight, self.nparts, self.nprofs, self.nbins)
        rec = libtomo.project(np.zeros(self.recreated.shape), flat_points,
                              weight, nUseParts, self.nprofs, self.nbins)
        # rec = self._normalize_profiles(rec)
        return rec

    # Convert x coordinates pointing at bins of flattened version of waterfall.
    def _create_flat_points(self) -> np.ndarray:
        return np.ascontiguousarray(
            super()._create_flat_points()).astype(np.int32)

    def _run_old(self, niter: int = 20, verbose: bool = False) -> np.ndarray:
        """Function to perform tomographic reconstruction.

        Performs the full reconstruction using C++.

        The discrepancy of each iteration is saved in the *diff* variable
        of the object.

        Parameters
        ----------
        niter: int
            Number of iterations in reconstruction.
        verbose: boolean
            Flag to indicate that the status of the tomography should be
            written to stdout. The output is identical to output
            generated in the original Fortran tomography.

        Returns
        -------
        weight: ndarray
            1D array containing the weight of each particle.

        Raises
        ------
        CoordinateError: Exception
            X-coordinates is None
        """
        if self.xp is None:
            raise expt.CoordinateError(
                'x-coordinates has value None, and must be provided')

        weight = np.ascontiguousarray(
            np.zeros(self.nparts, dtype=np.float64))
        self.diff = np.zeros(niter + 1, dtype=np.float64)

        flat_profiles = np.ascontiguousarray(
            self.waterfall.flatten().astype(np.float64))

        (self.weight,
         self.diff) = libtomo.reconstruct_old(
            weight, self.xp, flat_profiles,
            self.diff, niter, self.nbins,
            self.nparts, self.nprofs, verbose)
        return self.weight

    def run(self, niter: int = 20, verbose: bool = False,
            callback: t.Callable = None) -> np.ndarray:
        """Function to perform tomographic reconstruction.

        Performs the full reconstruction using C++.

        - The discrepancy of each iteration is saved in the objects\
          **diff** variable.
        - The particles weights are saved in the objects **weight** variable.
        - The reconstructed profiles are saved to the objects\
          **recreated** variable.

        Parameters
        ----------
        niter: int
            Number of iterations in reconstruction.
        verbose: boolean
            Flag to indicate that the status of the tomography should be
            written to stdout. The output is identical to output
            generated in the original Fortran tomography.
        callback: Callable
            Passing a callback with function signature
            (progress: int, total: int) will allow the tracking loop to call
            this function at the end of each turn, allowing the python caller
            to monitor the progress.

        Returns
        -------
        weight: ndarray
            1D array containing the weight of each particle.

        Raises
        ------
        CoordinateError: Exception
            X-coordinates is None
        """
        if self.xp is None:
            raise expt.CoordinateError(
                'x-coordinates has value None, and must be provided')

        (self.weight,
         self.diff,
         self.recreated) = conf.reconstruct(
            self.xp, self.waterfall, niter, self.nbins,
            self.nparts, self.nprofs, verbose, callback)
        return self.weight


# for backwards compatibility
TomographyCpp = Tomography



"""Unit-tests for the PhaseSpaceInfo class.

Run as python test_phase_space_info.py in console or via coverage
"""
from __future__ import annotations
import unittest

from .. import commons
import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.tracking.phase_space_info as psi
from longitudinal_tomography import exceptions as expt

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestMachine(unittest.TestCase):

    def test_find_dEbin_demax_lt_zero_vrf2_zero_output_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.xorigin = -246.60492626420734
        dEbin = psinfo.find_dEbin()

        correct_dEbin = 1232.7181430465346
        self.assertAlmostEqual(dEbin, correct_dEbin,
                               msg='dEbin was calculated incorrectly')

    def test_find_dEbin_demax_lt_zero_vrf2_not_zero_output_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.vrf2 = 1500
        machine.h_num = machine.vrf2 = 1500
        machine.h_num = 2
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.xorigin = -290.3024631321037
        dEbin = psinfo.find_dEbin()

        correct_dEbin = 185.1099466310996
        self.assertAlmostEqual(dEbin, correct_dEbin,
                               msg='dEbin was calculated incorrectly')

    def test_find_dEbin_demax_gt_zero_vrf2_not_zero_output_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.demax = 1.0e7
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.xorigin = -246.60492626420734
        dEbin = psinfo.find_dEbin()

        correct_dEbin = 26315.78947368421
        self.assertAlmostEqual(dEbin, correct_dEbin,
                               msg='dEbin was calculated incorrectly')

    def test_find_dEbin_demax_eq_zero_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.demax = 0.0
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.xorigin = -246.60492626420734

        with self.assertRaises(
                expt.EnergyBinningError,
                msg='demax equal to zero should raise exception'):
            dEbin = psinfo.find_dEbin()

    def test_find_dEbin_demax_xorigin_in_none_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.demax = 0.0
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)

        with self.assertRaises(
                expt.EnergyBinningError,
                msg='xorigin with value None should raise exception.'):
            dEbin = psinfo.find_dEbin()

    def test_calc_xorigin_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)
        xorigin = psinfo.calc_xorigin()

        correct_xorigin = -246.60492626420734
        self.assertAlmostEqual(xorigin, correct_xorigin,
                               msg='xorigin was calculated incorrectly')

    def test_find_binned_phase_energy_limits_correct_ilims(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        correct_imin = 0
        correct_imax = 759

        self.assertEqual(psinfo.imin, correct_imin,
                         msg='imin was calculated incorrectly'
                             '(full_pp_flag not enabled)')
        self.assertEqual(psinfo.imax, correct_imax,
                         msg='imin was calculated incorrectly'
                             '(full_pp_flag not enabled)')

    def test_find_binned_phase_energy_limits_correct_jmax(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        correct = [38, 38, 38, 38, 39, 45, 48, 50, 52, 54, 55, 57, 58, 59,
                   60, 61, 61, 62, 63, 63, 64, 65, 65, 65, 66, 66, 66, 67,
                   67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 67, 66, 66,
                   66, 66, 65, 65, 65, 64, 64, 63, 63, 62, 62, 61, 60, 60,
                   59, 58, 58, 57, 56, 56, 55, 54, 53, 52, 51, 50, 49, 48,
                   47, 46, 45, 43, 41, 38]

        for j, corr in zip(psinfo.jmax, correct):
            self.assertEqual(j, corr,
                             msg='jmax calculated incorrectly'
                                 '(full_pp_flag not enabled)')

    def test_find_binned_phase_energy_limits_correct_jmin(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        correct = [38, 38, 38, 38, 37, 31, 28, 26, 24, 22, 21, 19, 18, 17,
                   16, 15, 15, 14, 13, 13, 12, 11, 11, 11, 10, 10, 10, 9,
                   9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 11,
                   11, 11, 12, 12, 13, 13, 14, 14, 15, 16, 16, 17, 18, 18,
                   19, 20, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
                   33, 35, 38]

        for j, corr in zip(psinfo.jmin, correct):
            self.assertEqual(j, corr,
                             msg='jmin calculated incorrectly'
                                 '(full_pp_flag not enabled)')

    def test_find_binned_phase_energy_limits_full_pp_correct_ilims(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        machine.full_pp_flag = True

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        correct_imin = 0
        correct_imax = 759

        self.assertEqual(psinfo.imin, correct_imin,
                         msg='imin was calculated incorrectly'
                             '(full_pp_flag enabled)')
        self.assertEqual(psinfo.imax, correct_imax,
                         msg='imin was calculated incorrectly'
                             '(full_pp_flag enabled)')

    def test_find_binned_phase_energy_limits_full_pp_correct_jmax(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        machine.full_pp_flag = True

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        correct = [76] * machine.nbins

        for j, corr in zip(psinfo.jmax, correct):
            self.assertEqual(j, corr,
                             msg='jmax calculated incorrectly'
                                 '(full_pp_flag enabled)')

    def test_find_binned_phase_energy_limits_full_pp_correct_jmin(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        machine.full_pp_flag = True

        rbn = 10
        machine.nbins = int(machine.nbins / rbn)
        machine.dtbin *= rbn
        machine.synch_part_x /= float(rbn)

        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        correct = [1] * machine.nbins

        for j, corr in zip(psinfo.jmin, correct):
            self.assertEqual(j, corr,
                             msg='jmin calculated incorrectly'
                                 '(full_pp_flag enabled)')



import os

import matplotlib.pyplot as plt
import numpy as np

import longitudinal_tomography.utils.tomo_input as tin
import longitudinal_tomography

longitudinal_tomography.use_gpu()


def generate_bunch(bunch_position, bunch_length,
                   bunch_energy, energy_spread,
                   n_macroparticles):
    # Generating phase and energy arrays
    phase_array = np.linspace(bunch_position - bunch_length / 2,
                              bunch_position + bunch_length / 2,
                              100)

    energy_array = np.linspace(bunch_energy - energy_spread / 2,
                               bunch_energy + energy_spread / 2,
                               100)

    # Getting Hamiltonian on a grid
    phase_grid, deltaE_grid = np.meshgrid(phase_array, energy_array)

    # Bin sizes
    bin_phase = phase_array[1] - phase_array[0]
    bin_energy = energy_array[1] - energy_array[0]

    # Density grid, isodensity lines
    isodensity = ((phase_grid - bunch_position) / bunch_length * 2) ** 2. + \
                 ((deltaE_grid - bunch_energy) / energy_spread * 2) ** 2.
    density_grid = 1 - isodensity ** 2.
    density_grid[density_grid < 0] = 0
    density_grid /= np.sum(density_grid)

    # Generating particles randomly inside the grid cells according to the
    # provided density_grid
    indexes = np.random.choice(np.arange(0, np.size(density_grid)),
                               n_macroparticles, p=density_grid.flatten())

    # Randomize particles inside each grid cell (uniform distribution)
    particle_phase = np.ascontiguousarray(
        phase_grid.flatten()[indexes] + (np.random.rand(
            n_macroparticles) - 0.5) * bin_phase)
    particle_energy = np.ascontiguousarray(
        deltaE_grid.flatten()[indexes] + (np.random.rand(
            n_macroparticles) - 0.5) * bin_energy)

    return particle_phase, particle_energy


def drift(dphi, dE, hnum, beta, E0, eta):
    return dphi - 2 * np.pi * hnum * eta * dE / (beta ** 2 * E0)


def kick(dphi, dE, charge, voltage, E0):
    return dE + charge * voltage * np.sin(dphi) - E0


ex_dir = os.path.split(os.path.realpath(os.path.dirname(__file__)))[0]
in_file_pth = os.path.join(ex_dir, 'input_files', 'C500MidPhaseNoise.dat')

file = []
with open(in_file_pth, 'r') as f:
    for i in range(98):
        file.append(f.readline().strip())

machine, frame = tin.txt_input_to_machine(file)
machine.values_at_turns()

bunch_position = machine.synch_part_x * machine.dtbin * 0.0
bunch_length = machine.nbins * machine.dtbin
bunch_energy = 0.0
energy_spread = 1.0E6
n_parts = int(4E4)

dphi, denergy = generate_bunch(
    bunch_position, bunch_length, bunch_energy,
    energy_spread, n_parts)

dphi = dphi * np.pi / (machine.nbins * machine.dtbin)

all_dphi = []
all_denergy = []
for i in range(1000):
    dphi = drift(dphi, denergy, machine.h_num, machine.beta0[0],
                 machine.e_rest, machine.eta0[0])

    denergy = kick(dphi, denergy, machine.q, 1000, 0)
    if i % 10 == 0:
        all_dphi.append(dphi.tolist())
        all_denergy.append(denergy.tolist())
all_dphi = np.array(all_dphi)
all_denergy = np.array(all_denergy)

xorigin = bunch_position - np.min(all_dphi[0])
xp = all_dphi / np.pi * machine.nbins - xorigin

dEmax = np.max(all_denergy)
yp = all_denergy / dEmax * machine.nbins + machine.nbins / 2

for x, y in zip(xp[::10], yp[::10]):
    plt.scatter(x, y, s=0.5)
    plt.show()

# Filter particles
# transpose particles
# do tomography

# xorigin plot
# ------------
# x = np.linspace(-np.pi/2, np.pi/2, 50)
# plt.plot(x, np.sin(x))
# plt.plot([x[0], x[-1]], [0, 0], color='black')
# plt.plot([np.min(all_dphi[0]), np.min(all_dphi[0])], [-1, 1], color='g')
# plt.show()

# Particle trajectories
# ---------------------
# ipts = [0, 20, 200, 2000, 20000]
# plt.plot(all_dphi[:,ipts], all_denergy[:,ipts])
# plt.show()



import os
import numpy as np
from warnings import warn

class AppConfig:
    _precision = np.float64
    _gpu_enabled = False
    _active_dict = {}

    @classmethod
    def __update_active_dict(cls, new_dict):
        for key in cls._active_dict.keys():
            if key in globals():
                del globals()[key]
        globals().update(new_dict)
        cls._active_dict = new_dict


    @classmethod
    def get_precision(cls):
        return cls._precision

    @classmethod
    def set_single_precision(cls):
        cls._precision = np.float32
        if cls._gpu_enabled:
            cls.load_modules_and_refresh_kernels(True)

    @classmethod
    def set_double_precision(cls):
        cls._precision = np.float64
        if cls._gpu_enabled:
            cls.load_modules_and_refresh_kernels(False)

    @classmethod
    def load_modules_and_refresh_kernels(cls, single_prec_flag=False):
        cls.compile_kernels() # Check if compilation needed
        gpu_dev = GPUDev.get_gpu_dev()
        if single_prec_flag:
            gpu_dev.load_single_precision_modules()
        else:
            gpu_dev.load_double_precision_modules()
        from longitudinal_tomography.python_routines import kick_and_drift_cuda, reconstruct_cuda
        kick_and_drift_cuda.refresh_kernels()
        reconstruct_cuda.refresh_kernels()

    @classmethod
    def is_gpu_enabled(cls):
        return cls._gpu_enabled

    @classmethod
    def use_cpu(cls):
        """
        Use the CPU to perform the calculations.
        The precision of the functions is not set here.
        It will be automatically inferred when calling the libtomo functions.
        """
        from longitudinal_tomography.cpp_routines import libtomo
        cpu_func_dict = {
            'kick_and_drift': libtomo.kick_and_drift,
            'reconstruct': libtomo.reconstruct,
            'make_phase_space': libtomo.make_phase_space,
            'device': 'CPU'
        }

        for fname in dir(np):
            if callable(getattr(np, fname)) and (fname not in cpu_func_dict) \
                    and (fname[0] != '_'):
                cpu_func_dict[fname] = getattr(np, fname)

        cls.__update_active_dict(cpu_func_dict)
        cls._gpu_enabled = False


    @classmethod
    def use_gpu(cls, gpu_id=0):
        """
        Use the GPU device to perform the calculations
        
        Args:
            gpu_id (int, optional): Device id, default = 0
        """
        if gpu_id < 0:
            return

        import cupy as cp

        single_prec = True if cls._precision == np.float32 else False
        GPUDev(single_prec, gpu_id)

        from longitudinal_tomography.python_routines import kick_and_drift_cuda, reconstruct_cuda, data_treatment

        gpu_func_dict = {
            'kick_and_drift': kick_and_drift_cuda.kick_and_drift_cuda,
            'reconstruct': reconstruct_cuda.reconstruct_cuda,
            'make_phase_space': data_treatment.make_phase_space,
            'device': 'GPU'
        }

        for fname in dir(cp):
            if callable(getattr(cp, fname)) and (fname not in gpu_func_dict):
                gpu_func_dict[fname] = getattr(cp, fname)
        cls.__update_active_dict(gpu_func_dict)
        cls._gpu_enabled = True
    
    @classmethod
    def compile_kernels(cls, force=False):
        """
        Compile CUDA kernels if they have not been compiled yet or if forced.

        Args:
            force (bool, optional): If True, forces the compilation of CUDA kernels
                regardless whether they have already been compiled. Defaults to False.
        """
        from longitudinal_tomography import cuda_kernels
        if force or not cuda_kernels.check_compiled_kernels():
            if not force:
                warn("No compiled CUDA kernels found. Compiling CUDA kernels...")
            cuda_kernels.compile_kernels()


class GPUDev:
    __instance = None

    @classmethod
    def get_gpu_dev(cls):
        if cls.__instance is None:
            cls.__instance = GPUDev()
        return cls.__instance

    def __init__(self, single_prec = False, _gpu_id=0):
        if GPUDev.__instance is not None:
            return
        else:
            GPUDev.__instance = self

        import cupy as cp
        self.id = _gpu_id
        self.dev = cp.cuda.Device(self.id)
        self.dev.use()

        self.name = cp.cuda.runtime.getDeviceProperties(self.dev)['name']
        self.attributes = self.dev.attributes
        self.properties = cp.cuda.runtime.getDeviceProperties(self.dev)

        self.func_dict = {}

        # set the default grid and block sizes
        default_blocks = 2 * self.attributes['MultiProcessorCount']
        default_threads = self.attributes['MaxThreadsPerBlock']
        blocks = int(os.environ.get('GPU_BLOCKS', default_blocks))
        threads = int(os.environ.get('GPU_THREADS', default_threads))
        self.grid_size = (blocks, 1, 1)
        self.block_size = (threads, 1, 1)

        self.directory = os.path.dirname(os.path.realpath(__file__)) + "/"

        ## Compile if needed
        AppConfig.compile_kernels()

        if single_prec:
            self.load_single_precision_modules()
        else:
            self.load_double_precision_modules()

    def load_double_precision_modules(self):
        import cupy as cp
        self.kd_mod = cp.RawModule(path=os.path.join(
                    self.directory, f'../cuda_kernels/kick_and_drift_double.cubin'))
        self.rec_mod = cp.RawModule(path=os.path.join(
                            self.directory, f'../cuda_kernels/reconstruct_double.cubin'))

    def load_single_precision_modules(self):
        import cupy as cp
        self.kd_mod = cp.RawModule(path=os.path.join(
                    self.directory, f'../cuda_kernels/kick_and_drift_single.cubin'))
        self.rec_mod = cp.RawModule(path=os.path.join(
                            self.directory, f'../cuda_kernels/reconstruct_single.cubin'))

    def report_attributes(self):
        # Saves into a file all the device attributes
        with open(f'{self.name}-attributes.txt', 'w') as f:
            for k, v in self.attributes.items():
                f.write(f"{k}:{v}\n")

def cast(arr):
    """
    Cast an array (only floats) to a CuPy array if GPU is enabled, otherwise to a NumPy array

    Args:
        arr (numpy.ndarray or cupy.ndarray): The input array to be cast.

    Returns:
        numpy.ndarray or cupy.ndarray: The input array casted on the current device.
    """
    return cast_to_gpu(arr) if AppConfig.is_gpu_enabled() else cast_to_cpu(arr)

def cast_to_gpu(arr):
    """
    Cast an array (only floats) to GPU

    Args:
        arr (numpy.ndarray or cupy.ndarray): The input array to be cast to a CuPy array.

    Returns:
        cupy.ndarray: The input array on GPU.
    """
    import cupy as cp
    arr = cp.array(arr, dtype=AppConfig.get_precision())
    return arr

def cast_to_cpu(arr):
    """
    Cast an array (only floats) to CPU

    Args:
        arr (numpy.ndarray or cupy.ndarray): The input array to be cast to a NumPy memory.

    Returns:
        numpy.ndarray: The input array on CPU.
    """
    if hasattr(arr, 'get'):
        arr = arr.get().astype(AppConfig.get_precision())
    else:
        arr = np.array(arr, dtype=AppConfig.get_precision())
    return arr

AppConfig.use_cpu()


"""Unit-tests for the compat.fortran module.

Run as python test.py in console or via coverage
"""
from __future__ import annotations
import os
import shutil
import unittest

import numpy as np
import numpy.testing as nptest

from .. import commons
import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.tracking.particles as pts
from longitudinal_tomography import exceptions as expt
import longitudinal_tomography.compat.fortran as ftn

base_dir = os.path.split(os.path.realpath(__file__))[0]
base_dir = os.path.split(base_dir)[0]
tmp_dir = os.path.join(base_dir, 'tmp')

# All values retrieved from INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestCompat(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if not os.path.isdir(tmp_dir):
            os.mkdir(tmp_dir)

    @classmethod
    def tearDownClass(cls):
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)

    def test_calc_baseline_ftn_correct(self):
        waterfall = self._load_waterfall()
        correct = 0.00014048793859649817
        ans = ftn.calc_baseline(waterfall, ref_prof=0)
        self.assertAlmostEqual(
            ans, correct, msg='Baseline calculated incorrectly')

    def test_save_profile(self):
        waterfall = np.arange(25).reshape((5, 5))

        recprof = 2
        ftn.save_profile(waterfall, 2, tmp_dir)

        profile_pth = os.path.join(tmp_dir, f'profile{recprof + 1:03d}.data')
        with open(profile_pth, 'r') as f:
            read = f.readlines()

        correct = [' 1.0000000E+01\n', ' 1.1000000E+01\n', ' 1.2000000E+01\n',
                   ' 1.3000000E+01\n', ' 1.4000000E+01\n']

        nptest.assert_equal(
            read, correct, err_msg='Format of Fortran styled profiles file '
                                   'is incorrect')

    def test_save_vself(self):
        self_volts = np.arange(9).reshape((3, 3))

        ftn.save_self_volt_profile(self_volts, tmp_dir)

        profile_pth = os.path.join(tmp_dir, 'vself.data')
        with open(profile_pth, 'r') as f:
            read = f.readlines()

        correct = [' 0.0000000E+00\n', ' 1.0000000E+00\n', ' 2.0000000E+00\n',
                   ' 3.0000000E+00\n', ' 4.0000000E+00\n', ' 5.0000000E+00\n',
                   ' 6.0000000E+00\n', ' 7.0000000E+00\n', ' 8.0000000E+00\n']

        nptest.assert_equal(
            read, correct, err_msg='Format of Fortran styled vself file '
                                   'is incorrect')

    def test_save_phase_space(self):
        image = np.arange(9).reshape((3, 3))
        recprof = 2

        ftn.save_phase_space(image, recprof, tmp_dir)

        image_pth = os.path.join(tmp_dir, f'image{recprof + 1:03d}.data')
        with open(image_pth, 'r') as f:
            read = f.readlines()

        correct = ['  0.0000000E+00\n', '  1.0000000E+00\n',
                   '  2.0000000E+00\n', '  3.0000000E+00\n',
                   '  4.0000000E+00\n', '  5.0000000E+00\n',
                   '  6.0000000E+00\n', '  7.0000000E+00\n',
                   '  8.0000000E+00\n']

        nptest.assert_equal(
            read, correct, err_msg='Format of Fortran styled image file '
                                   'is incorrect')

    def test_save_difference(self):
        diff = np.arange(11) * np.pi / 100
        recprof = 2
        ftn.save_difference(diff, tmp_dir, recprof)

        diff_pth = os.path.join(tmp_dir, f'd{recprof + 1:03d}.data')
        with open(diff_pth, 'r') as f:
            read = f.readlines()

        correct = ['           0  0.0000000E+00\n',
                   '           1  3.1415927E-02\n',
                   '           2  6.2831853E-02\n',
                   '           3  9.4247780E-02\n',
                   '           4  1.2566371E-01\n',
                   '           5  1.5707963E-01\n',
                   '           6  1.8849556E-01\n',
                   '           7  2.1991149E-01\n',
                   '           8  2.5132741E-01\n',
                   '           9  2.8274334E-01\n',
                   '          10  3.1415927E-01\n']

        nptest.assert_equal(read, correct,
                            err_msg='Discrepancies was saved incorrectly.')

    def test_write_plotinfo(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()
        particles = pts.Particles()
        particles.homogeneous_distribution(machine, 0)
        profile_charge = 1e6

        plotinfo = ftn.write_plotinfo(machine, particles, profile_charge)
        plotinfo_list = np.array(plotinfo.split('\n'))

        correct = [' plotinfo.data',
                   'Number of profiles used in each reconstruction,',
                   ' profilecount = 150',
                   'Width (in pixels) of each image = length (in bins) '
                   'of each profile,',
                   ' profilelength = 760',
                   'Width (in s) of each pixel = width of each profile bin,',
                   ' dtbin = 1.0000E-09',
                   'Height (in eV) of each pixel,',
                   ' dEbin = 1.2327E+03',
                   'Number of elementary charges in each image,',
                   ' eperimage = 1.000E+06',
                   'Position (in pixels) of the reference synchronous point:',
                   ' xat0 =  334.000',
                   ' yat0 =  380.000',
                   'Foot tangent fit results (in bins):',
                   ' tangentfootl =    0.000',
                   ' tangentfootu =    0.000',
                   ' fit xat0 =   0.000',
                   'Synchronous phase (in radians):',
                   ' phi0( 1) = 0.4008',
                   'Horizontal range (in pixels) of the region in '
                   'phase space of map elements:',
                   ' imin( 1) =   0 and imax( 1) =  759']

        nptest.assert_equal(plotinfo_list, correct,
                            err_msg='Error in creation of plotinfo string')

    def test_write_plotinfo_fitted_x(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.values_at_turns()

        fittedx = 23.123445677
        tanbin_low = 3.12345567
        tanbin_up = 80.12345567
        machine.load_fitted_synch_part_x_ftn((fittedx, tanbin_low, tanbin_up))

        particles = pts.Particles()
        particles.homogeneous_distribution(machine, 0)
        profile_charge = 1e6

        plotinfo = ftn.write_plotinfo(machine, particles, profile_charge)
        plotinfo_list = np.array(plotinfo.split('\n'))

        correct = [' plotinfo.data',
                   'Number of profiles used in each reconstruction,',
                   ' profilecount = 150',
                   'Width (in pixels) of each image = length (in bins) '
                   'of each profile,',
                   ' profilelength = 760',
                   'Width (in s) of each pixel = width of each profile bin,',
                   ' dtbin = 1.0000E-09',
                   'Height (in eV) of each pixel,',
                   ' dEbin = 1.2327E+03',
                   'Number of elementary charges in each image,',
                   ' eperimage = 1.000E+06',
                   'Position (in pixels) of the reference synchronous point:',
                   ' xat0 =  23.123',
                   ' yat0 =  380.000',
                   'Foot tangent fit results (in bins):',
                   ' tangentfootl =    3.123',
                   ' tangentfootu =    80.123',
                   ' fit xat0 =   23.123',
                   'Synchronous phase (in radians):',
                   ' phi0( 1) = 0.4008',
                   'Horizontal range (in pixels) of the region in '
                   'phase space of map elements:',
                   ' imin( 1) =   0 and imax( 1) =  759']

        print(plotinfo_list)

        nptest.assert_equal(plotinfo_list, correct,
                            err_msg='Error in creation of plotinfo string. '
                                    'fitted value for synch. part. x '
                                    'was provided.')

    def test_write_plotinfo_no_dEbin_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        particles = pts.Particles()
        particles.imin = 0
        particles.imax = 100
        profile_charge = 1e6

        with self.assertRaises(expt.EnergyBinningError,
                               msg='An exception should be raised if the '
                                   'particles.dEbin is None'):
            ftn.write_plotinfo(machine, particles, profile_charge)

    def test_write_plotinfo_no_imin_imax_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        particles = pts.Particles()
        particles.imin = None
        particles.imax = 100
        particles.dEbin = 1e3
        profile_charge = 1e6

        with self.assertRaises(expt.PhaseLimitsError,
                               msg='An exception should be raised if '
                                   'particles.imin is None'):
            ftn.write_plotinfo(machine, particles, profile_charge)

        particles.imin = 0
        particles.imax = None

        with self.assertRaises(expt.PhaseLimitsError,
                               msg='An exception should be raised if '
                                   'particles.imax is None'):
            ftn.write_plotinfo(machine, particles, profile_charge)

    def test_load_fitted_synch_part_x_correct(self):
        new_x = 118.37008249622234
        bunch_min = 21.56907195264529
        bunch_max = 255.84863660530607
        fit_info = (new_x, bunch_min, bunch_max)

        machine = mch.Machine(**MACHINE_ARGS)
        machine.load_fitted_synch_part_x_ftn(fit_info)

        self.assertAlmostEqual(
            machine.fitted_synch_part_x, 118.37008249622234,
            msg='fitted synch x coordinate set incorrectly')
        self.assertAlmostEqual(
            machine.synch_part_x, 118.37008249622234,
            msg='fitted synch x coordinate set incorrectly')
        self.assertAlmostEqual(
            machine.bunchlimit_low, 21.56907195264529,
            msg='fitted lower bunch limit set incorrectly')
        self.assertAlmostEqual(
            machine.bunchlimit_up, 255.84863660530607,
            msg='fitted upper bunch limit set incorrectly')

    def _load_waterfall(self):
        base_dir = os.path.split(os.path.realpath(__file__))[0]
        base_dir = os.path.split(base_dir)[0]
        data_path = os.path.join(base_dir, 'resources')

        waterfall = np.load(os.path.join(
            data_path, 'waterfall_INDIVShavingC325.npy'))
        return waterfall



"""Unit-tests for the tomo_output module.

Run as python test_tomo_output.py in console or via coverage
"""

import os
import shutil
import unittest
from unittest.mock import patch

import numpy as np
import numpy.testing as nptest

from .. import commons
import longitudinal_tomography.utils.tomo_output as tout
import longitudinal_tomography.data.data_treatment as dtreat

base_dir = os.path.split(os.path.realpath(__file__))[0]
base_dir = os.path.split(base_dir)[0]
tmp_dir = os.path.join(base_dir, 'tmp')


class TestTomoOut(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if not os.path.isdir(tmp_dir):
            os.mkdir(tmp_dir)

    @classmethod
    def tearDownClass(cls):
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)

    def test_create_phase_space_image(self):

        weights = np.ones(25)

        ax = np.arange(5)
        xp, yp = np.meshgrid(ax, ax)
        xp = np.vstack(xp.flatten())
        yp = np.vstack(yp.flatten())

        nbins = 5
        recprof = 0
        img = tout.create_phase_space_image(xp, yp, weights, nbins, recprof)

        correct = np.array([[0.04, 0.04, 0.04, 0.04, 0.04],
                            [0.04, 0.04, 0.04, 0.04, 0.04],
                            [0.04, 0.04, 0.04, 0.04, 0.04],
                            [0.04, 0.04, 0.04, 0.04, 0.04],
                            [0.04, 0.04, 0.04, 0.04, 0.04]])

        nptest.assert_equal(
            img, correct, err_msg='Phase space image was created incorrectly')

    @patch('matplotlib.pyplot.show')
    def test_show(self, mock_show):
        _, _, profiles = commons.load_data()
        tomo, machine = commons.get_tomography_params()
        waterfall = profiles.waterfall
        rec_prof = 0

        phase_space = dtreat.phase_space(tomo, machine, rec_prof)[-1]
        measured_profile = waterfall[:, 0] / waterfall[:, 0].sum()

        tout.show(phase_space, measured_profile, tomo.diff)



"""
This module is deprecated. Its contents have been moved to
longitudinal_tomography.data.data_treatment.
This module exists solely for backwards compatibility and can be removed
without notice.
"""

from ..data.data_treatment import *
from warnings import warn

# warn('longitudinal_tomography.utils.data_treatment has been moved to '
#      'longitudinal_tomography.data.data_treatment')



"""Unit-tests for the tomo_input module.

Run as python test_tomo_input.py in console or via coverage
"""

import os
import unittest

import numpy as np
import numpy.testing as nptest

from .. import commons
import longitudinal_tomography.tracking.machine as mch
from longitudinal_tomography import exceptions as expt
import longitudinal_tomography.utils.tomo_input as tomoin

# All values retrieved from INDIVShavingC325.dat
frame_input_args = {
    'raw_data_path':       '',
    'framecount':          150,
    'skip_frames':         0,
    'framelength':         1000,
    'dtbin':               9.999999999999999E-10,
    'skip_bins_start':     170,
    'skip_bins_end':       70,
    'rebin':               3
}

MACHINE_ARGS = commons.get_machine_args()


class TestTomoIn(unittest.TestCase):

    # Tests for frame class
    # ---------------------

    def test_frames_set_raw_data_fails(self):
        frames = tomoin.Frames(**frame_input_args)

        with self.assertRaises(expt.RawDataImportError,
                               msg='raw data provided as scalar should raise '
                                   'an exception'):
            frames.raw_data = 1

        with self.assertRaises(expt.RawDataImportError,
                               msg='Wrong amount of data points should raise '
                                   'an exception'):
            frames.raw_data = [1]

    def test_nprofs_correct(self):
        frames = tomoin.Frames(**frame_input_args)
        nprof = frames.nprofs()
        self.assertEqual(nprof, 150, msg='Error in calculation of nprofs')

    def test_nbins_correct(self):
        frames = tomoin.Frames(**frame_input_args)
        nbins = frames.nbins()
        self.assertEqual(nbins, 760, msg='Error in calculation of nbins')

    def test_to_waterfall_skip_bins_both_sides(self):
        nframes = 3
        nbins_frame = 30
        dtbin = 9.999999999999999E-10
        skip_frames = 1
        skip_bins_start = 2
        skip_bins_end = 2
        rebin = 1

        ndata = nframes * nbins_frame

        raw_data = np.arange(ndata)

        frames = tomoin.Frames(
            nframes, nbins_frame, skip_frames, skip_bins_start,
            skip_bins_end, rebin, dtbin)

        waterfall = frames.to_waterfall(raw_data)

        correct = np.array([[32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43,
                             44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55,
                             56, 57],
                            [62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73,
                             74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85,
                             86, 87]])
        nptest.assert_equal(
            waterfall, correct,
            err_msg='error in conversion from raw data to waterfall')

    def test_to_waterfall_skip_start_bins(self):
        nframes = 3
        nbins_frame = 30
        dtbin = 9.999999999999999E-10
        skip_frames = 1
        skip_bins_start = 6
        skip_bins_end = 0
        rebin = 1

        ndata = nframes * nbins_frame

        raw_data = np.arange(ndata)

        frames = tomoin.Frames(
            nframes, nbins_frame, skip_frames, skip_bins_start,
            skip_bins_end, rebin, dtbin)

        waterfall = frames.to_waterfall(raw_data)

        correct = np.array([[36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
                             48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59],
                            [66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77,
                             78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89]])

        nptest.assert_equal(
            waterfall, correct,
            err_msg='error in conversion from raw data to waterfall')

    def test_to_waterfall_wrong_number_raw_data(self):
        nframes = 3
        nbins_frame = 30
        dtbin = 9.999999999999999E-10
        skip_frames = 1
        skip_bins_start = 6
        skip_bins_end = 0
        rebin = 1

        frames = tomoin.Frames(
            nframes, nbins_frame, skip_frames, skip_bins_start,
            skip_bins_end, rebin, dtbin)

        raw_data = np.arange(20)

        with self.assertRaises(expt.RawDataImportError,
                               msg='Wrong size of raw data relative to the '
                                   'given parameters should '
                                   'raise an exception'):
            frames.to_waterfall(raw_data)

    def test_to_waterfall_raw_data_no_iter(self):
        nframes = 3
        nbins_frame = 30
        dtbin = 9.999999999999999E-10
        skip_frames = 1
        skip_bins_start = 6
        skip_bins_end = 0
        rebin = 1

        frames = tomoin.Frames(
            nframes, nbins_frame, skip_frames, skip_bins_start,
            skip_bins_end, rebin, dtbin)

        raw_data = 1
        with self.assertRaises(expt.RawDataImportError,
                               msg='Raw data as scalar should raise '
                                   'an exception'):
            frames.to_waterfall(raw_data)

    # Tests for module functions
    # --------------------------

    def test_raw_data_to_profiles_no_iter_fails(self):
        machine = mch.Machine(**MACHINE_ARGS)
        waterfall = 1
        rbn = 1
        sampling_time = machine.dtbin

        with self.assertRaises(expt.WaterfallError,
                               msg='Waterfall as non iterable should raise '
                                   'an exception'):
            tomoin.raw_data_to_profiles(waterfall, machine, rbn, sampling_time)

    def test_raw_data_to_profiles_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        raw_data = self._load_raw_data()

        waterfall = raw_data.reshape((150, 1000))
        waterfall = waterfall[:, 170: -70]
        rbn = 3
        s_time = machine.dtbin

        profiles = tomoin.raw_data_to_profiles(
            waterfall, machine, rbn, s_time)

        correct_waterfall = self._load_waterfall()

        # Checking that waterfall was rebinned correctly
        nptest.assert_almost_equal(
            profiles.waterfall, correct_waterfall,
            err_msg='An found in the profiles.waterfall, after converting'
                    'from waterfall of raw data to profiles')

        # Checking if machine.dtbin was updated correctly
        correct_dtbin_e9 = 2.9999999999999996

        self.assertAlmostEqual(
            machine.dtbin * 1e9, correct_dtbin_e9,
            msg='The value of machine.dtbin not correct after rebinning.')

        # Checking that machine.synch_part_x was updated correctly
        correct_synch_part_x = 111.33333333333336
        self.assertAlmostEqual(
            machine.synch_part_x, correct_synch_part_x,
            msg='The value of machine.synch_part_x not '
                'correct after rebinning.')

        # Checking that the sampling time set in the profiles object
        # is correct.
        self.assertEqual(
            profiles.sampling_time, s_time,
            msg='The profile objects sampling time should have '
                'the value of the original dtbin')

    def test_txt_input_to_machine_machine_correct(self):
        data_path = self._find_resource_dir()
        infile = os.path.join(data_path, 'parameters_INDIVShavingC325.dat')

        with open(infile, 'r') as f:
            file_lines = f.readlines()
        machine, _ = tomoin.txt_input_to_machine(file_lines)

        self.assertEqual(machine.demax, MACHINE_ARGS['demax'],
                         msg='demax set incorrectly')

        self.assertEqual(machine.dturns, MACHINE_ARGS['dturns'],
                         msg='dturns set incorrectly')

        self.assertEqual(machine.vrf1, MACHINE_ARGS['vrf1'],
                         msg='vrf1 set incorrectly')

        self.assertEqual(machine.vrf1dot, MACHINE_ARGS['vrf1dot'],
                         msg='vrf1dot set incorrectly')

        self.assertEqual(machine.vrf2, MACHINE_ARGS['vrf2'],
                         msg='vrf2 set incorrectly')

        self.assertEqual(machine.vrf2dot, MACHINE_ARGS['vrf2dot'],
                         msg='vrf2dot set incorrectly')

        self.assertEqual(machine.mean_orbit_rad,
                         MACHINE_ARGS['mean_orbit_rad'],
                         msg='mean_orbit_rad set incorrectly')

        self.assertEqual(machine.bending_rad, MACHINE_ARGS['bending_rad'],
                         msg='bending_rad set incorrectly')

        self.assertEqual(machine.b0, MACHINE_ARGS['b0'],
                         msg='b0 set incorrectly')

        self.assertEqual(machine.bdot, MACHINE_ARGS['bdot'],
                         msg='bdot set incorrectly')

        self.assertEqual(machine.phi12, MACHINE_ARGS['phi12'],
                         msg='phi12 set incorrectly')

        self.assertEqual(machine.h_ratio, MACHINE_ARGS['h_ratio'],
                         msg='h_ratio set incorrectly')

        self.assertEqual(machine.trans_gamma, MACHINE_ARGS['trans_gamma'],
                         msg='trans_gamma set incorrectly')

        self.assertEqual(machine.e_rest, MACHINE_ARGS['rest_energy'],
                         msg='rest_energy set incorrectly')

        self.assertEqual(machine.q, MACHINE_ARGS['charge'],
                         msg='q set incorrectly')

        self.assertEqual(machine.g_coupling, MACHINE_ARGS['g_coupling'],
                         msg='g_coupling set incorrectly')

        self.assertEqual(machine.zwall_over_n, MACHINE_ARGS['zwall_over_n'],
                         msg='zwall_over_n set incorrectly')

        self.assertEqual(machine.min_dt, MACHINE_ARGS['min_dt'],
                         msg='min_dt set incorrectly')

        self.assertEqual(machine.max_dt, MACHINE_ARGS['max_dt'],
                         msg='max_dt set incorrectly')

        self.assertEqual(machine.pickup_sensitivity,
                         MACHINE_ARGS['pickup_sensitivity'],
                         msg='pickup_sensitivity set incorrectly')

        self.assertEqual(machine.synch_part_x, MACHINE_ARGS['synch_part_x'],
                         msg='synch_part_x set incorrectly')

        self.assertEqual(machine.dtbin, MACHINE_ARGS['dtbin'],
                         msg='dtbin set incorrectly')

        self.assertEqual(machine.self_field_flag,
                         MACHINE_ARGS['self_field_flag'],
                         msg='self_field_flag set incorrectly')

        self.assertEqual(machine.full_pp_flag,
                         MACHINE_ARGS['full_pp_flag'],
                         msg='full_pp_flag set incorrectly')

        self.assertEqual(machine.machine_ref_frame,
                         MACHINE_ARGS['machine_ref_frame'],
                         msg='machine_ref_frame set incorrectly')

        self.assertEqual(machine.beam_ref_frame,
                         MACHINE_ARGS['beam_ref_frame'],
                         msg='beam_ref_frame set incorrectly')

        self.assertEqual(machine.snpt, MACHINE_ARGS['snpt'],
                         msg='snpt set incorrectly')
        self.assertEqual(machine.niter, MACHINE_ARGS['niter'],
                         msg='niter set incorrectly')
        self.assertEqual(machine.filmstart, MACHINE_ARGS['filmstart'],
                         msg='filmstart set incorrectly')
        self.assertEqual(machine.filmstop, MACHINE_ARGS['filmstop'],
                         msg='filmstop set incorrectly')
        self.assertEqual(machine.filmstep, MACHINE_ARGS['filmstep'],
                         msg='filmstep set incorrectly')

        nprofiles = 150
        self.assertEqual(machine.nprofiles, nprofiles,
                         msg='nprofiles set incorrectly')
        nbins = 760
        self.assertEqual(machine.nbins, nbins,
                         msg='nprofiles set incorrectly')

    def test_txt_input_to_machine_frames_correct(self):
        data_path = self._find_resource_dir()
        infile = os.path.join(data_path, 'parameters_INDIVShavingC325.dat')

        with open(infile, 'r') as f:
            file_lines = f.readlines()
        _, frames = tomoin.txt_input_to_machine(file_lines)

        self.assertEqual(frames.nframes, frame_input_args['framecount'],
                         msg='The number of frames is not set correctly')

        self.assertEqual(frames.nbins_frame, frame_input_args['framelength'],
                         msg='The number of frame bins is not set correctly')

        self.assertEqual(frames.skip_frames, frame_input_args['skip_frames'],
                         msg='The number of frames to be skipped '
                             'is not set correctly')

        self.assertEqual(frames.skip_bins_start,
                         frame_input_args['skip_bins_start'],
                         msg='The number binst to be skipped in the start '
                             'of each frame is not set correctly')

        self.assertEqual(frames.skip_bins_end,
                         frame_input_args['skip_bins_end'],
                         msg='The number binst to be skipped in the end '
                             'of each frame is not set correctly')

        self.assertEqual(frames.rebin,
                         frame_input_args['rebin'],
                         msg='The rebin factor is set incorrectly')

        self.assertEqual(frames.sampling_time,
                         frame_input_args['dtbin'],
                         msg='The sampling time factor is set incorrectly')

    def test_txt_input_to_machine_no_iter_fails(self):
        not_iterable = 1

        with self.assertRaises(expt.InputError,
                               msg='non iterable input should '
                                   'raise an exception'):
            _, _ = tomoin.txt_input_to_machine(not_iterable)

    def test_txt_input_to_machine_wrong_len_fails(self):
        wrong_length_iter = [1, 2, 3]

        with self.assertRaises(expt.InputError,
                               msg='non iterable of the wrong length '
                                   'should raise an exception'):
            _, _ = tomoin.txt_input_to_machine(wrong_length_iter)

    def _load_waterfall(self):
        data_path = self._find_resource_dir()
        waterfall = np.load(os.path.join(
            data_path, 'waterfall_INDIVShavingC325.npy'))
        return waterfall

    def _load_raw_data(self):
        data_path = self._find_resource_dir()
        raw_data = np.genfromtxt(
            os.path.join(data_path,
                         'raw_data_INDIVShavingC325.dat'))
        return raw_data

    def _find_resource_dir(self):
        base_dir = os.path.split(os.path.realpath(__file__))[0]
        base_dir = os.path.split(base_dir)[0]
        data_path = os.path.join(base_dir, 'resources')
        return data_path



"""Unit-tests for the physics module.

Run as python test_physics.py in console or via coverage
"""

import unittest

import numpy as np

from .. import commons
import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.utils.physics as physics

# Machine arguments based on the input file INDIVShavingC325.dat
MACHINE_ARGS = commons.get_machine_args()


class TestPhysics(unittest.TestCase):

    def test_b_to_e_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        correct = 1015458784.835785
        ans = physics.b_to_e(machine)
        self.assertAlmostEqual(
            ans, correct, msg='Error in energy calculated from B field')

    def test_lorenz_beta_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.e0 = [1015458784.835785]
        correct = 0.382420087611783
        ans = physics.lorentz_beta(machine, rf_turn=0)
        self.assertAlmostEqual(
            ans, correct, msg='Error in calculation of Lorentz beta')

    def test_rfvolt_rf1_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.time_at_turn = [1.3701195948153858e-06]
        correct = -705.3144383117857
        ans = physics.rfvolt_rf1_mch(0.123, machine, rf_turn=0)

        self.assertAlmostEqual(
            float(ans), correct, msg='Error in calculation of rfvolt1')

    def test_drfvolt_rf1_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.time_at_turn = [1.3701195948153858e-06]
        correct = 2617.2730921111274
        ans = physics.drfvolt_rf1_mch(0.123, machine, rf_turn=0)

        self.assertAlmostEqual(
            float(ans), correct, msg='Error in calculation drfvolt1')

    def test_rf_voltage_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.vrf2 = 1250
        machine.time_at_turn = [1.3701195948153858e-06]
        correct = -1364.5929048626685
        ans = physics.rf_voltage_mch(0.123, machine, rf_turn=0)

        self.assertAlmostEqual(
            ans, correct, msg='Error in calculation rf_voltage')

    def test_drf_voltage_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.vrf2 = 1250
        machine.time_at_turn = [1.3701195948153858e-06]
        correct = 4741.280534228331
        ans = physics.drf_voltage_mch(0.123, machine, rf_turn=0)

        self.assertAlmostEqual(
            ans, correct, msg='Error in calculation drf_voltage')

    def test_rf_voltage_correct2(self):
        machine = mch.Machine(**MACHINE_ARGS)
        correct = -1132.2371121228516
        ans = physics.rf_voltage_at_phase(
            0.123, 2500, 0, 2135, 0,
            2, 1.324, [1.3701195948],
            0)

        self.assertAlmostEqual(
            ans, correct, msg='Error in calculation drf_voltage')

    def test_vrft_correct(self):
        correct = 1500
        ans = physics.vrft(500, 250, 4)
        self.assertEqual(ans, correct, msg='Error in calculation vrft')

    def test_find_synchronous_phase_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.time_at_turn = [1.3701195948153858e-06]
        machine.e0 = [1015458784.835785]

        correct = 2.740810528223139
        ans = physics.find_synch_phase_mch(
            machine, rf_turn=0, phi_lower=0, phi_upper=2*np.pi)

        self.assertAlmostEqual(
            ans, correct, msg='Error in found synchronous phase')

    def test_find_phi_lower_upper_positive_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.e0 = [1015458784.835785]
        correct_low = -np.pi
        correct_up = np.pi
        ans_low, ans_up = physics.find_phi_lower_upper(machine, rf_turn=0)

        self.assertAlmostEqual(
            ans_low, correct_low, msg='Error in found calculation '
                                      'of lower phase')
        self.assertAlmostEqual(
            ans_up, correct_up, msg='Error in found calculation '
                                    'of upper phase')

    def test_find_phi_lower_upper_negative_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.e0 = [1015458784.835785]
        machine.q *= -1
        correct_low = 0.0
        correct_up = 2 * np.pi
        ans_low, ans_up = physics.find_phi_lower_upper(machine, rf_turn=0)

        self.assertAlmostEqual(
            ans_low, correct_low, msg='Error in found calculation '
                                      'of lower phase')
        self.assertAlmostEqual(
            ans_up, correct_up, msg='Error in found calculation '
                                    'of upper phase')

    def test_phase_slip_factor_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.beta0 = np.array([0.38242009])
        correct = 0.7942664750023455
        ans = physics.phase_slip_factor(machine)
        ans = ans[0]
        self.assertAlmostEqual(
            ans, correct, msg='Error in found calculation phase slip factor')

    def test_find_dphase_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.beta0 = np.array([0.38242009])
        machine.eta0 = np.array([0.79426648])
        machine.e0 = np.array([1015458784.835785])

        correct = 3.360488420030597e-08
        ans = physics.find_dphase(machine)
        ans = ans[0]

        self.assertAlmostEqual(
            ans, correct, msg='Error in found calculation dphase')

    def test_calc_revolution_freq_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.beta0 = np.array([0.38242009])

        correct = 4585866.350787248
        ans = physics.revolution_freq(machine)
        ans = ans[0]
        self.assertAlmostEqual(
            ans, correct, msg='Error in found calculation of '
                              'revolution frequency')

    def test_calc_self_field_coeffs_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.beta0 = np.array([0.38242009])
        machine.omega_rev0 = np.array([4585866.32214847])
        machine.nprofiles = 1
        machine.dturns = 1
        machine.g_coupling = 1.0
        machine.zwall_over_n = 50.0

        correct = 1.2945175200855999e-05
        ans = physics.calc_self_field_coeffs(machine)
        ans = ans[0]

        self.assertAlmostEqual(
            ans, correct, msg='Error in found calculation of '
                              'self-field coefficients')

    def test_phase_low_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.phi0 = [0.40078213]
        machine.time_at_turn = [1.3701195948153858e-06]
        bunch_phaselength = 3.22312485522767
        correct = -1883.316491755506
        ans = physics.phase_low_mch(0.123, machine, bunch_phaselength,
                                    rf_turn=0)

        self.assertAlmostEqual(ans, correct,
                               msg='Phase_low calculated incorrectly')

    def test_dphase_low_correct(self):
        machine = mch.Machine(**MACHINE_ARGS)
        machine.phi0 = [0.40078213]
        machine.time_at_turn = [1.3701195948153858e-06]
        bunch_phaselength = 3.22312485522767
        correct = 859.1967476555656
        ans = physics.dphase_low_mch(0.123, machine, bunch_phaselength, 0,
                                     None)

        self.assertAlmostEqual(ans, correct,
                               msg='dphase_low calculated incorrectly')



import subprocess
import os
from .. import exceptions as expt

REDUCTION_BLOCK_SIZE = 32

cuda_sources = [
    os.path.dirname(__file__) + "/kick_and_drift",
    os.path.dirname(__file__) + "/reconstruct",
]

def compile_kernels():
    """
    Compiles the CUDA source files to binaries which will be imported.
    If it cannot find nvcc naturally, please set the environment variable CUDA_PATH
    to be able to call the compiler.
    """

    print("Compiling the CUDA sources")

    cuda_path = os.getenv("CUDA_PATH", default="")
    if cuda_path != "":
        nvcc = cuda_path + "/bin/nvcc"
    else:
        nvcc = "nvcc"

    nvccflags = [nvcc, "--cubin", "-O3", "--use_fast_math"]

    try:
        import cupy as cp
        dev = cp.cuda.Device(0)
    except ImportError as e:
        raise expt.CudaCompilationException(
            "Package CuPy is not installed. CUDA sources cannot be compiled"
        ) from e
    except cp.cuda.runtime.CUDARuntimeError as e:
        raise expt.CudaCompilationException(
            "No capable GPU device has been found.\
                CUDA sources cannot be compiled"
        ) from e
    else:
        dev_name = cp.cuda.runtime.getDeviceProperties(dev)["name"]
        print("Discovering the device compute capability..")
        comp_capability = dev.compute_capability
        print(f"Device name {dev_name}")

        print(f"Compiling the sources for architecture {comp_capability}.")
        nvccflags += ["-arch", f"sm_{comp_capability}"]

        try:
            for source in cuda_sources:
                command = nvccflags + [
                    "-o",
                    source + "_double.cubin",
                    source + ".cu",
                    f"-DBLOCK_SIZE={REDUCTION_BLOCK_SIZE}"
                ]
                subprocess.run(command)
                command = nvccflags + [
                    "-o",
                    source + "_single.cubin",
                    source + ".cu",
                    "-DUSEFLOAT",
                    f"-DBLOCK_SIZE={REDUCTION_BLOCK_SIZE}"
                ]
                subprocess.run(command)
        except FileNotFoundError as e:
            raise expt.CudaCompilationException(
                "The NVCC compiler could not be found. "\
                + "Please check if CUDA_PATH has been set properly. "\
                + f"Subprocess returned the following error: {e}"
            )

        if check_compiled_kernels():
            print("The CUDA sources have been successfully compiled.")
        else:
            raise expt.CudaCompilationException(
                "The CUDA source compilation failed."
        )

def check_compiled_kernels():
    """
    Check if the compiled CUDA kernels for both double and single precision exist.

    Returns:
        bool: True if all compiled CUDA kernels exist, False otherwise.
    """
    return os.path.isfile(cuda_sources[0] + "_double.cubin")\
        and os.path.isfile(cuda_sources[1] + "_double.cubin")\
        and os.path.isfile(cuda_sources[0] + "_single.cubin")\
        and os.path.isfile(cuda_sources[1] + "_single.cubin")


import os

import matplotlib.pyplot as plt
import numpy as np

import longitudinal_tomography.tomography.tomography as tomography
import longitudinal_tomography.tracking.particles as parts
import longitudinal_tomography.tracking.tracking as tracking
import longitudinal_tomography.utils.tomo_input as tomoin
import longitudinal_tomography.utils.tomo_config

import longitudinal_tomography
longitudinal_tomography.use_gpu()
longitudinal_tomography.set_double_precision()

if os.getenv("DATAFILE") is not None:
    datafile = os.getenv("DATAFILE")
else:
    datafile = "flatTopINDIVRotate2.dat"


ex_dir = os.path.split(os.path.realpath(os.path.dirname(__file__)))[0]
in_file_pth = os.path.join(ex_dir, "input_files", datafile)
parameter_lines = 98

input_parameters = []
with open(in_file_pth, "r") as line:
    for i in range(parameter_lines):
        input_parameters.append(line.readline().strip())

raw_data = np.genfromtxt(in_file_pth, skip_header=98)
machine, frames = tomoin.txt_input_to_machine(input_parameters)
measured_waterfall = frames.to_waterfall(raw_data)

ntest_points = 20
vary_rfv = 1000  # volts
rfv_start = machine.vrf1 - vary_rfv
rfv_end = machine.vrf1 + vary_rfv
rfv_inputs = np.linspace(rfv_start, rfv_end, ntest_points)

profiles = tomoin.raw_data_to_profiles(
    measured_waterfall, machine, frames.rebin, frames.sampling_time
)

tomo = tomography.Tomography(profiles.waterfall)


diffs = []
for rfv in rfv_inputs:
    print(f"Reconstructing using rf-voltage of {rfv:.3f} volt")
    machine.vrf1 = rfv
    machine.values_at_turns()

    tracker = tracking.Tracking(machine)
    xp, yp = tracker.track(machine.filmstart)

    xp, yp = parts.physical_to_coords(
        xp, yp, machine, tracker.particles.xorigin, tracker.particles.dEbin
    )
    xp, yp = parts.ready_for_tomography(xp, yp, machine.nbins)

    tomo.xp = xp
    weight = tomo.run(niter=machine.niter)
    tomo.xp = None

    diffs.append(tomo.diff[-1])

# plt.plot(rfv_inputs, diffs)
# ax = plt.gca()
# ax.set_xlabel("Input voltage")
# ax.set_ylabel("Discrepancy")
# ax.ticklabel_format(axis="y", scilimits=(0, 0), style="sci")
# plt.show()



"""Module containing phase space post-processing functions.

:Author(s): **Anton Lu**
"""
from __future__ import annotations
from typing import TYPE_CHECKING

import numpy as np
from scipy import constants as cont
from multipledispatch import dispatch
from .. import assertions as asrt
from numbers import Number

if TYPE_CHECKING:
    from typing import Union, Dict, Iterable

    FloatArr = np.ndarray[float]

__all__ = ['post_process', 'rms_dpp', 'emittance_rms',
           'emittance_90', 'emittance_fractional']

m_p = cont.value('proton mass energy equivalent in MeV') * 1e6


def post_process(phase_space: Iterable[float], t_bins: Iterable[float],
                 e_bins: Iterable[float], energy: float, mass: float = m_p) \
                                -> Dict[str, Union[float, FloatArr]]:
    """
    Convenience function that provides an all-on-one post-processing method.

    Calculates the RMS emittance, 90% emittance and RMS dp/p of the given
    phase space

    Parameters
    ----------
    phase_space : np.ndarray
        A 2-dimensional array that represents the phase space of size
        (#t_bins, #e_bins)
    t_bins : np.ndarray
        A 1 or 2-dimensional vector that represent the bins on the time axis
    e_bins : np.ndarray
        A 1 or 2-dimensional vector that represents the bins on the energy axis
    energy : Number
        The energy of the beam at the reconstructed profile
    mass : Number
        The rest mass of the particle(s)

    Returns
    -------
    dict :
        A dictionary with the keys
        - emittance_rms
        - emittance_90
        - rms_dp/p
        Where k/v pair is one beam quality measure

    """

    time_proj = phase_space.sum(1)
    energy_proj = phase_space.sum(0)
    t_std = _std_from_histogram(time_proj, t_bins)
    e_std = _std_from_histogram(energy_proj, e_bins)

    out = {
        'emittance_rms': emittance_rms(t_std, e_std),
        'emittance_90': emittance_90(phase_space, t_bins, e_bins),
        'rms_dp/p': rms_dpp(e_std, energy, mass)
    }

    return out

#TODO: How does static analysis handle @dispatch functions?
@dispatch(np.ndarray, np.ndarray, np.ndarray)
def emittance_rms(histogram: np.ndarray,
                  t_bins: np.ndarray,
                  e_bins: np.ndarray) -> Union[float, np.ndarray]:
    """
    Calculates the RMS emittance for the given phase space
    by first calculating the standard deviation of the time and energy
    histograms

    Parameters
    ----------
    histogram : np.ndarray
        A 2-dimensional array that represents the phase space of size
        (#t_bins, #e_bins)
    t_bins : np.ndarray
        A 1 or 2-dimensional vector that represent the bins on the time axis
    e_bins : np.ndarray
        A 1 or 2-dimensional vector that represents the bins on the energy axis

    Returns
    -------
    float :
        The calculated RMS emittance
    """
    t_std = _std_from_histogram(histogram.sum(axis=1), t_bins)
    e_std = _std_from_histogram(histogram.sum(axis=0), e_bins)

    return emittance_rms(t_std, e_std)


@dispatch(float, float)
def emittance_rms(sigma_t: Union[float, np.ndarray],
                  sigma_e: Union[float, np.ndarray]) \
        -> Union[float, np.ndarray]:
    """
    Calculates the RMS emittance from the standard deviation of
    time and energy.

    Parameters
    ----------
    sigma_t : float
        Standard deviation of the time axis
    sigma_e : float
        Standard deviation of the energy axis

    Returns
    -------
    float :
        The calculated RMS emittance
    """
    return np.pi * sigma_t * sigma_e


def emittance_90(phase_space: Iterable[float], t_bins: Iterable[float],
                 e_bins: Iterable[float]) -> float:
    """
    Convenience function to calculate the 90% emittance that calls
    the emittance_fractional function.

    Parameters
    ----------
    phase_space : np.ndarray
        A 2-dimensional array that represents the phase space
    t_bins : np.ndarray
        A 1 or 2-dimensional vector that represent the bins on the time axis
    e_bins : np.ndarray
        A 1 or 2-dimensional vector that represents the bins on the energy axis

    Returns
    -------
    float :
        The 90% emittance of the passed phase space

    """
    return emittance_fractional(phase_space, t_bins, e_bins, fraction=90)


@dispatch(np.ndarray, np.ndarray, np.ndarray, fraction=float)
def emittance_fractional(histogram: np.ndarray,
                         t_bins: np.ndarray,
                         e_bins: np.ndarray,
                         fraction: float = 90) -> float:
    """
    Calculates the fractional emittance of a given phase space (histogram),
    defaults to 90% emittance if the `fraction` argument is left out.

    Parameters
    ----------
    histogram : np.ndarray
        A 2-dimensional array that represents the phase space
    t_bins : np.ndarray
        A 1 or 2-dimensional vector that represent the bins on the time axis
    e_bins : np.ndarray
        A 1 or 2-dimensional vector that represents the bins on the energy axis
    fraction: float
        A number between 0 and 100 (not inclusive), calculates the fraction%
        emittance.

    Returns
    -------
    float :
        The fractional emittance of the passed phase space

    """
    # TODO: explore options for faster calculations
    asrt.assert_less_or_equal(fraction, 'fraction', 100, ValueError)
    asrt.assert_greater_or_equal(fraction, 'fraction', 0, ValueError)

    fraction /= 100

    histogram = histogram / histogram.sum()
    histogram = histogram.ravel()
    histogram[::-1].sort()
    cumsum = histogram.cumsum()
    n_bins = np.argmax(cumsum >= fraction * histogram.sum())

    t_bin_width = t_bins[1] - t_bins[0]
    e_bin_width = e_bins[1] - e_bins[0]

    return n_bins * (t_bin_width * e_bin_width)


@dispatch(Number, Number, Number)
def rms_dpp(energy_std: Number, energy: Number, mass: Number) -> float:
    """
    Calculates the RMS dp/p using the energy projection histogram.
    Provided as a convenience function that first calculates
    the standard deviation of the energy and then calculates the
    RMS dp/p using the std.

    dp is calculated using the relation ((p + dp)c)^2 + m^2c^4 = (E + dE)^2

    Parameters
    ----------
    energy_std : Number
        The standard deviation of the energy axis
    energy : Number
        The energy of the beam at the reconstructed profile
    momentum : Number
        The momentum of the beam at the reconstructed profile
    mass : Number
        The rest mass of the particle(s)

    Returns
    -------
    float :
        The RMS dp/p of particle beam
    """
    momentum = np.sqrt(energy ** 2 - mass ** 2)
    dp = np.sqrt(np.power(energy + energy_std, 2) - mass ** 2) - momentum

    rms_dpp = dp / momentum

    return rms_dpp


@dispatch(np.ndarray, np.ndarray, Number, Number)
def rms_dpp(energy_proj: np.ndarray, energy_bins: np.ndarray,
            energy: Number,
            mass: Number) -> float:
    """
    Calculates the RMS dp/p using the energy projection histogram.
    Provided as a convenience function that first calculates
    the standard deviation of the energy and then calculates the
    RMS dp/p using the std.

    Parameters
    ----------
    energy_proj : np.ndarray
        A 1 or 2-dimensional vector that represents the energy projection of
        the phase space
    energy_bins : np.ndarray
        A 1 or 2-dimensional vector that represents the bins on the energy axis
    energy : Number
        The energy of the beam at the reconstructed profile
    momentum : Number
        The momentum of the beam at the reconstructed profile
    mass : Number
        The rest mass of the particle(s)

    Returns
    -------
    float :
        The RMS dp/p of the energy projection
    """
    energy_std = _std_from_histogram(energy_proj, energy_bins)

    return rms_dpp(energy_std, energy, mass)


def _std_from_histogram(histogram: Iterable[float], bins: Iterable[float])\
                                                                      -> float:
    """
    Calculates the standard deviation of a given histogram and the bin values

    Parameters
    ----------
    histogram : np.ndarray
        A histogram to calculate the standard deviation for
    bins : np.ndarray
        The bins of the histogram. Dimensions must match that of the histogram

    Returns
    -------
    float :
        The standard deviation of the histogram
    """
    asrt.assert_array_numel_equal((histogram, bins), ('histogram', 'bins'),
                                  histogram.size)

    # guard against 2 dim arrays
    bins = bins.flatten()
    histogram = histogram.flatten()

    bin_width = bins[1] - bins[0]
    mid_values = bins + bin_width / 2

    mean = mid_values * histogram / histogram.sum()
    var = np.sum(histogram * np.power(mid_values - mean, 2)) / histogram.sum()

    std = np.sqrt(var)
    return std



"""
Example implementation to create benchmarks for the GPU implementation.
First, bunch profiles will be created with BLonD using different
parameters. The focus will be on parameters for the PS Booster.
Second, the tracking and the reconstruction will be executed.
The runtime will be measured using Pyprof and saved.
"""

from pyprof import timing
import time

start_time = time.time()

timing.start_timing('import_packages')

# General imports
import numpy as np
import matplotlib.pyplot as plt
import os

# BLonD imports
from blond.input_parameters.ring import Ring
from blond.input_parameters.rf_parameters import RFStation
from blond.trackers.tracker import RingAndRFTracker, FullRingAndRF
from blond.beam.beam import Beam, Proton
from blond.beam.distributions import matched_from_distribution_function
from blond.beam.profile import Profile, CutOptions
from blond.monitors.monitors import SlicesMonitor

# Tomography imports
from longitudinal_tomography.data import data_treatment as dtreat
import longitudinal_tomography.tracking.machine as mch
import longitudinal_tomography.tracking.particles as parts
import longitudinal_tomography.tracking.tracking as tracking
import longitudinal_tomography.tomography.tomography as tomography
import longitudinal_tomography.utils.tomo_output as tomoout
from longitudinal_tomography.utils import tomo_config as conf



# Constant imports
from scipy.constants import c, e, m_p

timing.stop_timing()

# SIMULATION PARAMETERS -------------------------------------------------------
timing.start_timing('blond::set_params')

# Beam parameters
n_particles = int(1e11) # ?
n_macroparticles = int(1e5) # only in BLonD

distribution_exponent = None
bunch_length_fit = 'full'
distribution_type = 'parabolic_line'
max_bunch_length = 0.5476677e-6
bunch_length = max_bunch_length * 0.6   # [s] between 10% and 90% of RF period (0.5476677e-6)

n_bins_arr = np.array([100])                           # Between 50 and 2000
n_bins = 100

# Machine and RF parameters
radius = 25.0                   # for PSB
bending_radius = 8.239          # for PSB
gamma_transition = 4.1          # for PSB
C = 2 * np.pi * radius          # [m]

# Tracking details
n_turns_arr = np.array([100]) # n_profiles!!!
n_turns = 100
d_turns_arr = np.array([9])
dturns = 9

# Derived parameters
E_0 = m_p * c**2 / e            # [eV]
E_kin = 2e9                     # [eV] FT: 2e9 (or 1.4e9), FB: 160e6
tot_beam_energy = E_0 + E_kin   # [eV]
sync_momentum = np.sqrt(tot_beam_energy**2 + E_0**2)    # [eV]
momentum_compaction = 1 / gamma_transition**2
charge = 1 # -1 if electron
b0 = sync_momentum / bending_radius / c                 # [T]

gamma = tot_beam_energy / E_0
beta = np.sqrt(1.0-1.0/gamma**2.0)

# Cavity parameters
n_rf_systems = 1
h = 1.0 # for PSB
voltage_program = 24e3 # for PSB
phi_offset = np.pi

timing.stop_timing()
# DEFINE BLonD OBJECTS --------------------------------------------------------

timing.start_timing("set_device")
if os.getenv('SINGLE_PREC') is not None:
    conf.AppConfig.set_single_precision() if os.getenv('SINGLE_PREC') == 'True' else conf.AppConfig.set_double_precision()

if os.getenv('MODE') is not None:
    if os.getenv('MODE') == "Numba":
        conf.AppConfig.use_numba()
    elif os.getenv('MODE') == "CPP":
        conf.AppConfig.use_cpu()
    elif os.getenv('MODE') == "CuPy":
        conf.AppConfig.use_cupy()
        timing.mode = timing.Mode.CUPY
    elif os.getenv('MODE') == "CUDA":
        conf.AppConfig.use_gpu()
        timing.mode = timing.Mode.CUPY
    else:
        print("No mode given, using CPP")
        conf.AppConfig.use_cpu()
timing.stop_timing()

timing.start_timing('blond:create_objects')
general_params = Ring(C, momentum_compaction, sync_momentum, Proton(),
                    n_turns, bending_radius=bending_radius)
RF_st_par = RFStation(general_params, [h], [voltage_program], [phi_offset],
                    n_rf_systems)
beam = Beam(general_params, n_macroparticles, n_particles)
ring_RF_section = RingAndRFTracker(RF_st_par, beam)
full_tracker = FullRingAndRF([ring_RF_section])

fs = RF_st_par.omega_s0[0]/2/np.pi
bucket_length = 2.0 * np.pi / RF_st_par.omega_rf[0,0]

slice_beam = Profile(beam, CutOptions(cut_left=0, cut_right=bucket_length, n_slices=n_bins))
monitor = SlicesMonitor('./blonddata', n_turns, slice_beam)

timing.stop_timing()

# BEAM GENERATION -------------------------------------------------------------

timing.start_timing('blond::match_and_track')

distr = matched_from_distribution_function(beam, full_tracker,
                                distribution_type=distribution_type,
                                distribution_exponent=distribution_exponent,
                                bunch_length=bunch_length,
                                bunch_length_fit=bunch_length_fit,
                                distribution_variable='Action', seed=18)

beam.dE *= 0.3

# BLonD Tracking

bunch_profiles = np.zeros((n_turns, n_bins))

for i in range(n_turns):
    full_tracker.track()
    slice_beam.track()
    bunch_profiles[i] = slice_beam.n_macroparticles
    # BUNCH parameter?
    monitor.track("")
monitor.close()
timing.stop_timing()

#import blond.plots.plot_beams as bpb

#bpb.plot_long_phase_space(general_params, RF_st_par, beam, 0, 547.8e-9, -65e6, 65e6, show_plot = True, separatrix_plot = True, histograms_plot = False)

# TOMOGRAPHY PROCESS ----------------------------------------------------------

# DEFINE MACHINE 

timing.start_timing('define_machine_object')

n_profiles = len(bunch_profiles)
dtbin = slice_beam.bin_centers[1] - slice_beam.bin_centers[0]

machine_args = {
    'output_dir':           '/tmp/',
    'dtbin':                dtbin,
    'dturns':               dturns,
    'synch_part_x':         n_bins // 2, #np.argmax(bunch_profiles[0]),
    'demax':                -1.E6,              # noqa
    'filmstart':            0,
    'filmstop':             1,
    'filmstep':             1,
    'niter':                20,
    'snpt':                 4, # Square root of particles pr. cell of phase space.
    'full_pp_flag':         False,
    'beam_ref_frame':       0,
    'machine_ref_frame':    0,
    'vrf1':                 voltage_program,
    'vrf1dot':              0.0,
    'vrf2':                 0.0,
    'vrf2dot':              0.0,
    'h_num':                h,
    'h_ratio':              2.0,
    'phi12':                phi_offset, #0.4007821253666541,
    'b0':                   b0,
    'bdot':                 0.0,
    'mean_orbit_rad':       radius,
    'bending_rad':          bending_radius,
    'trans_gamma':          gamma_transition,
    'rest_energy':          E_0,
    'charge':               charge,
    'self_field_flag':      False,
    'g_coupling':           0.0,
    'zwall_over_n':         0.0,
    'pickup_sensitivity':   0.36,
    'nprofiles':            n_profiles,
    'nbins':                n_bins,
    'min_dt':               0.0,
    'max_dt':               dtbin * n_bins
}
machine = mch.Machine(**machine_args)
timing.stop_timing()

# bunch_profiles = waterfall?
timing.start_timing('create_profile')
waterfall = conf.array(bunch_profiles)
timing.stop_timing()

ntest_points = 100
vary_rfv = 2000  # volts
rfv_start = machine.vrf1 - vary_rfv
rfv_end = machine.vrf1 + vary_rfv
rfv_inputs = np.linspace(rfv_start, rfv_end, ntest_points)

precisions = ["single", "double"]

end_time = time.time()
if os.getenv("REPORT_FILENAME") is not None and os.getenv("REPORT_FILENAME") != "":
        report_filename = os.getenv("REPORT_FILENAME")
        timing.report(total_time = (end_time - start_time) * 1e3, out_file=report_filename + f"-baseprog-Voltagecalib")
        timing.reset()
else:
    timing.report(total_time = (end_time - start_time) * 1e3)
    timing.reset()

for prec in precisions:
    if prec == "double":
        conf.AppConfig.set_double_precision()
    elif prec == "single":
        conf.AppConfig.set_single_precision()

#for n_bins in n_bins_arr:
#    for n_turns, dturns in zip(n_turns_arr, d_turns_arr):
    firstit = True
    diffs = []
    for rfv in rfv_inputs:
        if firstit:
            start_time = time.time()

        timing.start_timing('values_at_turns')
        machine.vrf1 = rfv
        machine.values_at_turns()
        timing.stop_timing()

        timing.start_timing('tracking::create_tracker')
        tracker = tracking.Tracking(machine)
        reconstruct_idx = machine.filmstart
        timing.stop_timing()

        xp, yp = tracker.track(reconstruct_idx)

        timing.start_timing("physical_to_coords")
        xp, yp = parts.physical_to_coords(
            xp, yp, machine, tracker.particles.xorigin,
            tracker.particles.dEbin)
        timing.stop_timing()

        timing.start_timing("ready_for_tomo")
        xp, yp = parts.ready_for_tomography(xp, yp, machine.nbins)
        timing.stop_timing()

        timing.start_timing("create_tomo_object")
        tomo = tomography.Tomography(waterfall, xp, yp)
        timing.stop_timing()
        weight = tomo.run(niter=machine.niter)
        timing.start_timing("set_tomo_xp_none")
        tomo.xp = None
        timing.stop_timing()

        diffs.append(tomo.diff[-1])

        if firstit:
            end_time = time.time()
            if os.getenv("REPORT_FILENAME") is not None and os.getenv("REPORT_FILENAME") != "":
                report_filename = os.getenv("REPORT_FILENAME")
                timing.report(total_time = (end_time - start_time) * 1e3, out_file=report_filename + f"-{prec}-{n_bins}-bins-{n_turns}-profs-Voltagecalib-it1")
                timing.reset()
            else:
                timing.report(total_time = (end_time - start_time) * 1e3)
                timing.reset()
            firstit = False
            start_time = time.time()

    end_time = time.time()
    if os.getenv("REPORT_FILENAME") is not None and os.getenv("REPORT_FILENAME") != "":
        report_filename = os.getenv("REPORT_FILENAME")
        timing.report(total_time = (end_time - start_time) * 1e3, out_file=report_filename + f"-{prec}-{n_bins}-bins-{n_turns}-profs-Voltagecalib-itrest")
        timing.reset()
    else:
        timing.report(total_time = (end_time - start_time) * 1e3)
        timing.reset()

# while reconstruct_idx < n_turns:
#     image = tomoout.create_phase_space_image(
#         xp, yp, weight, machine.nbins, reconstruct_idx, mode=mode)
#     image = cp.asnumpy(image)
#     waterfall = cp.asnumpy(waterfall)
#     tomoout.show(image, tomo.diff, waterfall[reconstruct_idx])
#     t_range, E_range, density = dtreat.phase_space(tomo, machine,
#                                                        reconstruct_idx, mode)
#     density = density.get()
#     vmin = np.min(density[density > 0])
#     vmax = np.max(density)
#     plt.contourf(t_range * 1E9, E_range / 1E6, density.T,
#              levels=np.linspace(vmin, vmax, 50), cmap='Oranges')
#     plt.xlabel('dt (ns)')
#     plt.ylabel('dE (MeV)')
#     plt.show()
#     reconstruct_idx += 100


import os

import numpy as np

import longitudinal_tomography.tomography.tomography as tomography
import longitudinal_tomography.tracking.particles as parts
import longitudinal_tomography.tracking.tracking as tracking
import longitudinal_tomography.data.data_treatment as dtreat
import longitudinal_tomography.utils.tomo_input as tomoin
import longitudinal_tomography.utils.tomo_output as tomoout

import longitudinal_tomography
longitudinal_tomography.use_gpu()


ex_dir = os.path.split(os.path.realpath(os.path.dirname(__file__)))[0]
in_file_pth = os.path.join(ex_dir, 'input_files', 'C500MidPhaseNoise.dat')

parameter_lines = 98
input_parameters = []
with open(in_file_pth, 'r') as line:
    for i in range(parameter_lines):
        input_parameters.append(line.readline().strip())

raw_data = np.genfromtxt(in_file_pth, skip_header=98, dtype=np.float32)

# Generating machine object
machine, frames = tomoin.txt_input_to_machine(input_parameters)
machine.values_at_turns()
measured_waterfall = frames.to_waterfall(raw_data)

# Creating profiles object
profiles = tomoin.raw_data_to_profiles(
    measured_waterfall, machine,
    frames.rebin, frames.sampling_time)

profiles.calc_profilecharge()

if profiles.machine.synch_part_x < 0:
    fit_info = dtreat.fit_synch_part_x(profiles)
    machine.load_fitted_synch_part_x_ftn(fit_info)

tracker = tracking.Tracking(machine)
tracker.enable_fortran_output(profiles.profile_charge)

# For including self fields during tracking
# FOR SELF FIELDS
if machine.self_field_flag:
    profiles.calc_self_fields()
    tracker.enable_self_fields(profiles)

for film in range(machine.filmstart, machine.filmstop, machine.filmstep):

    xp, yp = tracker.track(film)

    # Converting from physical coordinates ([rad], [eV])
    # to phase space coordinates.
    if not tracker.self_field_flag:
        xp, yp = parts.physical_to_coords(
            xp, yp, machine, tracker.particles.xorigin,
            tracker.particles.dEbin)

    # Filters out lost particles, transposes particle matrix, casts to np.int32
    xp, yp = parts.ready_for_tomography(xp, yp, machine.nbins)

    # Reconstructing phase space
    tomo = tomography.TomographyCpp(profiles.waterfall, xp)
    weight = tomo.run(niter=machine.niter, verbose=True)

    # Creating image for fortran style presentation of phase space.
    image = tomoout.create_phase_space_image(
        xp, yp, weight, machine.nbins, film)
    tomoout.show(image, tomo.diff, profiles.waterfall[film])



"""Unit-tests for the TomographyCpp class.

Run as python test_tomography_cpp.py in console or via coverage
"""
from __future__ import annotations
import os
import unittest

import numpy as np
import numpy.testing as nptest

import longitudinal_tomography.tomography.tomography as tmo
from longitudinal_tomography import exceptions as expt
from longitudinal_tomography.utils import tomo_config as conf

from .. import commons

SINGLE_PREC_ACCURACY_PLACES = 5


class TestTomographyCpp(unittest.TestCase):

    def setUp(self):
        conf.AppConfig.use_cpu()
        conf.AppConfig.set_double_precision()

    def test_set_xp_not_iterable_fails(self):
        waterfall = np.ones((10, 10))
        tomo = tmo.TomographyCpp(waterfall)

        with self.assertRaises(expt.CoordinateImportError,
                               msg='Providing x coordinates as non iterable '
                                   'should raise an Exception'):
            tomo.xp = 1

    def test_set_xp_one_dim_fails(self):
        waterfall = np.ones((10, 10))
        tomo = tmo.TomographyCpp(waterfall)

        with self.assertRaises(expt.CoordinateImportError,
                               msg='Providing x coordinates array having '
                                   'another number of dimensions than two '
                                   'should raise an Exception'):
            tomo.xp = [1, 2, 4]

    def test_set_xp_wrong_nprofs_fails(self):
        nprofs = 5
        nbins = 10
        waterfall = np.ones((nprofs, nbins))
        tomo = tmo.TomographyCpp(waterfall)

        nparts = 20
        pts_nprofs = 2
        with self.assertRaises(expt.CoordinateImportError,
                               msg='Providing x coordinates array which has '
                                   'been tracked trough another number of '
                                   'profiles than found in the waterfall '
                                   'should raise an Exception'):
            tomo.xp = np.ones((nparts, pts_nprofs))

    def test_set_xp_outside_of_img_upper_fails(self):
        nprofs = 5
        img_width = 10
        waterfall = np.ones((nprofs, img_width))
        tomo = tmo.TomographyCpp(waterfall)

        nparts = 20
        xp = np.ones((nparts, nprofs))
        xp[1, 2] = img_width
        with self.assertRaises(expt.XPOutOfImageWidthError,
                               msg='Providing x coordinates array outside '
                                   'of image width should raise an Exception'):
            tomo.xp = xp

    def test_set_xp_outside_of_img_lower_fails(self):
        nprofs = 5
        img_width = 10
        waterfall = np.ones((nprofs, img_width))
        tomo = tmo.TomographyCpp(waterfall)

        nparts = 20
        xp = np.ones((nparts, nprofs))
        xp[1, 2] = -1
        with self.assertRaises(expt.XPOutOfImageWidthError,
                               msg='Providing x coordinates array outside '
                                   'of image width should raise an Exception'):
            tomo.xp = xp

    def test_xp_can_be_set_to_none(self):
        nprofs = 5
        img_width = 10
        nparts = 20
        waterfall = np.ones((nprofs, img_width))
        xp = np.ones((nparts, nprofs))
        tomo = tmo.TomographyCpp(waterfall, xp)
        tomo.xp = None

        self.assertEqual(
            tomo.xp, None, msg='xp should be able to be set to None')

    def test_waterfall_reduced_to_zeros_fail(self):
        nprofs = 5
        img_width = 10
        waterfall = -np.ones((nprofs, img_width))

        with self.assertRaises(expt.WaterfallReducedToZero,
                               msg='Providing an all negative or '
                                   'zero waterfall should raise an Exception'):
            tomo = tmo.TomographyCpp(waterfall)

    def test_run_xp_is_none_fails(self):
        nprofs = 5
        img_width = 10
        waterfall = np.ones((nprofs, img_width))
        tomo = tmo.TomographyCpp(waterfall)

        with self.assertRaises(expt.CoordinateError,
                               msg='Calling the run function, with '
                                   'longitudinal_tomography.xp=None '
                                   'should raise an Exception'):
            tomo.run()

    def test_run_correct_weights(self):
        waterfall = commons.load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.TomographyCpp(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        cweights = np.array([0.00861438, 0.00880599, 0.00893367, 0.00909539,
                             0.00918903, 0.00922729, 0.00928258, 0.00928246,
                             0.00927822, 0.00929515, 0.00929504, 0.00921845,
                             0.0092014,  0.00918014, 0.00915893, 0.00914618,
                             0.00919728, 0.00912925, 0.00918038, 0.00924009,
                             0.00920603, 0.00930828, 0.00940621, 0.00953821,
                             0.00959356, 0.00969566, 0.00981058, 0.00988708,
                             0.00998924, 0.01007859, 0.01014255, 0.01023618,
                             0.01021914, 0.01028299, 0.01026588, 0.01021909,
                             0.01016806, 0.01015526, 0.01011276, 0.01008299,
                             0.00999782, 0.00992973, 0.00991267, 0.00985308,
                             0.00977646, 0.0097126,  0.00971697, 0.00967434,
                             0.00961899, 0.00961908])

        nptest.assert_almost_equal(
            weights, cweights, err_msg='Weights were calculated incorrectly')

    def test_run_correct_weights_singleprec(self):
        conf.AppConfig.set_single_precision()
        waterfall = commons.load_waterfall().astype(np.float32)

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.Tomography(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        cweights = np.array([0.00861438, 0.00880599, 0.00893367, 0.00909539,
                             0.00918903, 0.00922729, 0.00928258, 0.00928246,
                             0.00927822, 0.00929515, 0.00929504, 0.00921845,
                             0.0092014,  0.00918014, 0.00915893, 0.00914618,
                             0.00919728, 0.00912925, 0.00918038, 0.00924009,
                             0.00920603, 0.00930828, 0.00940621, 0.00953821,
                             0.00959356, 0.00969566, 0.00981058, 0.00988708,
                             0.00998924, 0.01007859, 0.01014255, 0.01023618,
                             0.01021914, 0.01028299, 0.01026588, 0.01021909,
                             0.01016806, 0.01015526, 0.01011276, 0.01008299,
                             0.00999782, 0.00992973, 0.00991267, 0.00985308,
                             0.00977646, 0.0097126,  0.00971697, 0.00967434,
                             0.00961899, 0.00961908])

        nptest.assert_almost_equal(
            weights, cweights, decimal=SINGLE_PREC_ACCURACY_PLACES, err_msg='Weights were calculated incorrectly')

    def test_run_correct_diff(self):
        waterfall = commons.load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.TomographyCpp(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        correct = 0.009561478717303546
        self.assertAlmostEqual(tomo.diff[0], correct,
                               msg='Discrepancy calculated incorrectly')

    def test_run_correct_diff_singleprec(self):
        conf.AppConfig.set_single_precision()
        waterfall = commons.load_waterfall().astype(np.float32)

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.TomographyCpp(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        correct = 0.009561478717303546
        self.assertAlmostEqual(tomo.diff[0], correct, places=SINGLE_PREC_ACCURACY_PLACES,
                               msg='Discrepancy calculated incorrectly')

    def test_run_hybrid_reduced_to_zeros_fails(self):
        waterfall = commons.load_waterfall()

        nprofs = waterfall.shape[0]
        nparts = 50
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.TomographyCpp(waterfall, xp.T)

        with self.assertRaises(expt.WaterfallReducedToZero,
                               msg='Waterfall getting reduced to zero '
                                   'should raise an Exception'):
            tomo.run_hybrid()

    def test_run_hybrid_xp_is_none_fails(self):
        nprofs = 5
        img_width = 10
        waterfall = np.ones((nprofs, img_width))
        tomo = tmo.TomographyCpp(waterfall)

        with self.assertRaises(expt.CoordinateError,
                               msg='Calling the run_hybrid function, with '
                                   'longitudinal_tomography.xp=None should '
                                   'raise an Exception'):
            tomo.run_hybrid()

    def test_run_hybrid_correct(self):
        waterfall = commons.load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.TomographyCpp(waterfall)
        tomo.xp = xp.T
        weights = tomo.run_hybrid(1)

        cweights = np.array([0.00861438, 0.00880599, 0.00893367, 0.00909539,
                             0.00918903, 0.00922729, 0.00928258, 0.00928246,
                             0.00927822, 0.00929515, 0.00929504, 0.00921845,
                             0.0092014,  0.00918014, 0.00915893, 0.00914618,
                             0.00919728, 0.00912925, 0.00918038, 0.00924009,
                             0.00920603, 0.00930828, 0.00940621, 0.00953821,
                             0.00959356, 0.00969566, 0.00981058, 0.00988708,
                             0.00998924, 0.01007859, 0.01014255, 0.01023618,
                             0.01021914, 0.01028299, 0.01026588, 0.01021909,
                             0.01016806, 0.01015526, 0.01011276, 0.01008299,
                             0.00999782, 0.00992973, 0.00991267, 0.00985308,
                             0.00977646, 0.0097126,  0.00971697, 0.00967434,
                             0.00961899, 0.00961908])

        nptest.assert_almost_equal(
            weights, cweights, err_msg='Weights were calculated incorrectly')

    def test_run_hybrid_correct_diff(self):
        waterfall = commons.load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.TomographyCpp(waterfall)
        tomo.xp = xp.T
        weights = tomo.run_hybrid(1)

        correct = 0.009561478717303548
        self.assertAlmostEqual(tomo.diff[0], correct,
                               msg='Discrepancy calculated incorrectly')

class TestTomographyGPU(unittest.TestCase):

    def setUp(self):
        try:
            import cupy as cp
        except ImportError:
            self.skipTest('CuPy not found - skipped tests')
        else:
            self.cp = cp
            conf.AppConfig.use_gpu()
            conf.AppConfig.set_double_precision()

    def tearDown(self) -> None:
        conf.AppConfig.use_cpu()
        conf.AppConfig.set_double_precision()


    def test_set_xp_not_iterable_fails(self):
        waterfall = self.cp.ones((10, 10))
        tomo = tmo.Tomography(waterfall)

        with self.assertRaises(expt.CoordinateImportError,
                               msg='Providing x coordinates as non iterable '
                                   'should raise an Exception'):
            tomo.xp = 1

    def test_set_xp_one_dim_fails(self):
        waterfall = self.cp.ones((10, 10))
        tomo = tmo.Tomography(waterfall)

        with self.assertRaises(expt.CoordinateImportError,
                               msg='Providing x coordinates array having '
                                   'another number of dimensions than two '
                                   'should raise an Exception'):
            tomo.xp = self.cp.array([1, 2, 4])

    def test_set_xp_wrong_nprofs_fails(self):
        nprofs = 5
        nbins = 10
        waterfall = self.cp.ones((nprofs, nbins))
        tomo = tmo.Tomography(waterfall)

        nparts = 20
        pts_nprofs = 2
        with self.assertRaises(expt.CoordinateImportError,
                               msg='Providing x coordinates array which has '
                                   'been tracked trough another number of '
                                   'profiles than found in the waterfall '
                                   'should raise an Exception'):
            tomo.xp = self.cp.ones((nparts, pts_nprofs))

    def test_set_xp_outside_of_img_upper_fails(self):
        nprofs = 5
        img_widthh = 10
        waterfall = self.cp.ones((nprofs, img_widthh))
        tomo = tmo.Tomography(waterfall)

        nparts = 20
        xp = self.cp.ones((nparts, nprofs))
        xp[1, 2] = img_widthh
        with self.assertRaises(expt.XPOutOfImageWidthError,
                               msg='Providing x coordinates array outside '
                                   'of image width should raise an Exception'):
            tomo.xp = xp

    def test_set_xp_outside_of_img_lower_fails(self):
        nprofs = 5
        img_width = 10
        waterfall = self.cp.ones((nprofs, img_width))
        tomo = tmo.Tomography(waterfall)

        nparts = 20
        xp = self.cp.ones((nparts, nprofs))
        xp[1, 2] = -1
        with self.assertRaises(expt.XPOutOfImageWidthError,
                               msg='Providing x coordinates array outside '
                                   'of image width should raise an Exception'):
            tomo.xp = xp

    def test_xp_can_be_set_to_none(self):
        nprofs = 5
        img_width = 10
        nparts = 20
        waterfall = self.cp.ones((nprofs, img_width))
        xp = self.cp.ones((nparts, nprofs))
        tomo = tmo.Tomography(waterfall, xp)
        tomo.xp = None

        self.assertEqual(
            tomo.xp, None, msg='xp should be able to be set to None')

    def test_waterfall_reduced_to_zeros_fail(self):
        nprofs = 5
        img_width = 10
        waterfall = -self.cp.ones((nprofs, img_width))

        with self.assertRaises(expt.WaterfallReducedToZero,
                               msg='Providing an all negative or '
                                   'zero waterfall should raise an Exception'):
            tomo = tmo.Tomography(waterfall)

    def test_run_xp_is_none_fails(self):
        nprofs = 5
        img_width = 10
        waterfall = self.cp.ones((nprofs, img_width))
        tomo = tmo.Tomography(waterfall)

        with self.assertRaises(expt.CoordinateError,
                               msg='Calling the run function, with '
                                   'longitudinal_tomography.xp=None '
                                   'should raise an Exception'):
            tomo.run()

    def test_run_correct_weights(self):
        waterfall = self.cp.array(commons.load_waterfall())

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = self.cp.meshgrid(self.cp.arange(0, nparts), self.cp.arange(nprofs))[0]

        tomo = tmo.Tomography(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        cweights = np.array([0.00861438, 0.00880599, 0.00893367, 0.00909539,
                             0.00918903, 0.00922729, 0.00928258, 0.00928246,
                             0.00927822, 0.00929515, 0.00929504, 0.00921845,
                             0.0092014,  0.00918014, 0.00915893, 0.00914618,
                             0.00919728, 0.00912925, 0.00918038, 0.00924009,
                             0.00920603, 0.00930828, 0.00940621, 0.00953821,
                             0.00959356, 0.00969566, 0.00981058, 0.00988708,
                             0.00998924, 0.01007859, 0.01014255, 0.01023618,
                             0.01021914, 0.01028299, 0.01026588, 0.01021909,
                             0.01016806, 0.01015526, 0.01011276, 0.01008299,
                             0.00999782, 0.00992973, 0.00991267, 0.00985308,
                             0.00977646, 0.0097126,  0.00971697, 0.00967434,
                             0.00961899, 0.00961908])

        self.cp.testing.assert_array_almost_equal(
            weights, cweights, err_msg='Weights were calculated incorrectly on GPU')

    def test_run_correct_weights_singleprec(self):
        conf.AppConfig.set_single_precision()
        waterfall = self.cp.array(commons.load_waterfall(), dtype=self.cp.float32)

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = self.cp.meshgrid(self.cp.arange(0, nparts), self.cp.arange(nprofs))[0]

        tomo = tmo.Tomography(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        cweights = np.array([0.00861438, 0.00880599, 0.00893367, 0.00909539,
                             0.00918903, 0.00922729, 0.00928258, 0.00928246,
                             0.00927822, 0.00929515, 0.00929504, 0.00921845,
                             0.0092014,  0.00918014, 0.00915893, 0.00914618,
                             0.00919728, 0.00912925, 0.00918038, 0.00924009,
                             0.00920603, 0.00930828, 0.00940621, 0.00953821,
                             0.00959356, 0.00969566, 0.00981058, 0.00988708,
                             0.00998924, 0.01007859, 0.01014255, 0.01023618,
                             0.01021914, 0.01028299, 0.01026588, 0.01021909,
                             0.01016806, 0.01015526, 0.01011276, 0.01008299,
                             0.00999782, 0.00992973, 0.00991267, 0.00985308,
                             0.00977646, 0.0097126,  0.00971697, 0.00967434,
                             0.00961899, 0.00961908])

        self.cp.testing.assert_array_almost_equal(
            weights, cweights, decimal=SINGLE_PREC_ACCURACY_PLACES, err_msg='Weights were calculated incorrectly on GPU')
        
    def test_run_gpu_cpu_equal(self):
        conf.AppConfig.use_cpu()
        waterfall = commons.load_waterfall()

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = np.meshgrid(np.arange(0, nparts), np.arange(nprofs))[0]

        tomo = tmo.Tomography(waterfall)
        tomo.xp = xp.T
        weights = tomo.run(1)

        conf.AppConfig.use_gpu()

        waterfall_gpu = self.cp.array(commons.load_waterfall())
        waterfall_gpu = waterfall_gpu[:nprofs]
        waterfall_gpu = waterfall_gpu[:, 70:170]
        xp_gpu = self.cp.meshgrid(self.cp.arange(0, nparts), self.cp.arange(nprofs))[0]

        tomo_gpu = tmo.Tomography(waterfall_gpu)
        tomo_gpu.xp = xp_gpu.T
        weights_gpu = tomo_gpu.run(1)

        self.cp.testing.assert_array_almost_equal(
            weights, weights_gpu, err_msg='CPU and GPU weights do not match')

        self.cp.testing.assert_array_almost_equal(
            tomo.diff[0], tomo_gpu.diff[0], err_msg='CPU and GPU difference do not match')

    def test_run_correct_diff(self):
        waterfall = self.cp.array(commons.load_waterfall())

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = self.cp.meshgrid(self.cp.arange(0, nparts), self.cp.arange(nprofs))[0]

        tomo = tmo.Tomography(waterfall)
        tomo.xp = xp.T
        tomo.run(1)

        correct = 0.009561478717303546
        self.cp.testing.assert_array_almost_equal(tomo.diff[0], correct,
                               err_msg='Discrepancy calculated incorrectly on GPU')

    def test_run_correct_diff_singleprec(self):
        conf.AppConfig.set_single_precision()
        waterfall = self.cp.array(commons.load_waterfall(), dtype=self.cp.float32)

        nprofs = 10
        nparts = 50
        waterfall = waterfall[:nprofs]
        waterfall = waterfall[:, 70:170]
        xp = self.cp.meshgrid(self.cp.arange(0, nparts), self.cp.arange(nprofs))[0]

        tomo = tmo.Tomography(waterfall)
        tomo.xp = xp.T
        tomo.run(1)

        correct = 0.009561478717303546
        self.cp.testing.assert_array_almost_equal(tomo.diff[0], correct, decimal=SINGLE_PREC_ACCURACY_PLACES,
                               err_msg='Discrepancy calculated incorrectly on GPU')


"""
Module containing the Particles class for creating and storing
and initial particle distribution.

Module also contains utility functions for particle distribution
like assertions, conversions and filtering of lost particles.

:Author(s): **Christoffer Hjertø Grindheim**, **Anton Lu**
"""
from __future__ import annotations
import logging
from typing import Tuple, Sequence, TYPE_CHECKING, Union
import numpy as np

from . import phase_space_info as psi
from .. import assertions as asrt, exceptions as expt
from ..utils import tomo_config as conf

if TYPE_CHECKING:
    from .machine import Machine
    from .machine_base import MachineABC

log = logging.getLogger(__name__)


class Particles(object):
    """Class holding the initial particle distribution to be tracked.

    An automatic homogeneous particle based on the algorithm from the original
    tomography program can be created, by calling
    :func:`homogeneous_distribution`.

    Attributes
    ----------
    dEbin: float, default=None
        Energy size of bins in phase space coordinate system.
    xorigin: float, default=None
        The absolute difference (in bins) between phase=0 and the origin
        of the reconstructed phase space coordinate system.
    imin: int, default=None
        Minimum phase of reconstruction area. Given in bins phase space.
    imax: int, default=None
        Maximum phase of reconstruction area. Given in bins phase space.
    jmin: ndarray, default=None
        Minimum energy for each phase of the phase space coordinate system.
    jmax: ndarray, default=None
        Maximum energy for each phase of the phase space coordinate system.
    coordinates_dphi_denergy: tuple, default=(None, None)
        Tuple containing coordinates of initial particle distribution as 1D
        ndarrays (dphi, denergy). These are the differences in phase [rad]
        energy [eV] relative to the synchronous particle.
    """

    def __init__(self):
        self.dEbin = None
        self.xorigin = None

        # Area where particles have been automatically populated.
        # I - bin in phase axis, J - bin in energy axis
        self.imin = None
        self.imax = None
        self.jmin = None
        self.jmax = None

        self._dphi = None
        self._denergy = None

    @property
    def coordinates_dphi_denergy(self) -> Tuple[np.ndarray, np.ndarray]:
        """Particle coordinates defined as @property.
        Use this property to provide and retrieve coordinates
        of the initial particle distribution in units of phase [rad]
        and energy [eV] relative to the synchronous particle.

        This property contains a built-in assertion of the provided particles.

        Returns
        -------
        coordinates: tuple
            Returns a tuple holding (dphi, denergy)

            * dphi
                ndarray containing the phase [rad] of each particle
                reative to the synchronous particle.
            * denergy
                ndarray containing the energy [eV] of each particle
                reative to the synchronous particle.
        """
        return self._dphi, self._denergy

    @coordinates_dphi_denergy.setter
    def coordinates_dphi_denergy(self,
                                 coordinates: Tuple[np.ndarray, np.ndarray]):
        """
        See the definition of the property
        :py:func`longitudinal_tomography.tracking.particles.
                 coordinates_dphi_denergy` for full documentation.

        Parameters
        ----------
        coordinates: tuple
            Tuple holding (dphi, denergy)

            * dphi
                ndarray containing the phase [rad] of each particle
                relative to the synchronous particle.
            * denergy
                ndarray containing the energy [eV] of each particle
                relative to the synchronous particle.
        """
        self._dphi, self._denergy = _assert_coordinates(coordinates)

    def homogeneous_distribution(self, machine: 'Machine', recprof: int,
                                 deltaturn: int = 0):
        """Function for automatic generation of particle distribution.

        The distributions created are identical to the distributions created
        in the Fortran tomography. The reconstruction area is found by calling
        :func:`~longitudinal_tomography.tracking.phase_space_info.
                PhaseSpaceInfo.find_binned_phase_energy_limits`.

        Parameters
        ----------
        machine: Machine
            Object containing the machine settings during the measurements.
            Needed for the calculation of the reconstruction area, and
            for setting the the number of particles to be created.
        recprof: int
            The index of the profile (time frame) to be reconstructed.
            This will be the profile where the distribution is generated.
        deltaturn: int
            On offset in turn numbers to allow creation of distributions
            between two recorded profiles

        Raises
        ------
        MachineParameterError: Exception
            Raised if fields are missing from provided machine object.
        """
        self._assert_machine(machine)

        # The reconstruction area is found
        psinfo = psi.PhaseSpaceInfo(machine)
        psinfo.find_binned_phase_energy_limits()

        self.dEbin = psinfo.dEbin
        machine.dEbin = self.dEbin
        self.xorigin = psinfo.xorigin

        nbins_y = np.sum(psinfo.jmax[psinfo.imin: psinfo.imax + 1]
                         - psinfo.jmin[psinfo.imin: psinfo.imax + 1])

        # Creating the distribution of particles within one cell of the phase
        # space coordinate system.
        bin_pts = ((2.0 * np.arange(1, machine.snpt + 1) - 1)
                   / (2.0 * machine.snpt))

        # Creating x coordinates
        x = np.arange(psinfo.imin, psinfo.imax + 1, dtype=float)
        nbins_x = len(x)
        x = np.repeat(x, machine.snpt)
        x += np.tile(bin_pts, nbins_x)

        # Creating y coordinates.
        nbins_y = np.max(psinfo.jmax) - np.min(psinfo.jmin)
        y = np.arange(np.min(psinfo.jmin), np.max(psinfo.jmax), dtype=float)
        y = np.repeat(y, machine.snpt)
        y += np.tile(bin_pts, nbins_y)

        coords = np.meshgrid(x, y)
        coords = np.array([coords[0].flatten(), coords[1].flatten()])

        # Remove particles outside of the limits of i (phase) and j (energy)
        # of the reconstruction area.
        coords = coords[:, coords[1] < psinfo.jmax[coords[0].astype(int)]]
        coords = coords[:, coords[1] > psinfo.jmin[coords[0].astype(int)]]

        self.imin = psinfo.imin
        self.imax = psinfo.imax
        self.jmin = psinfo.jmin
        self.jmax = psinfo.jmax

        # Converting from phase space coordinates to physical units.
        coords = conf.cast(self._bin_nr_to_physical_coords(coords, machine, recprof, deltaturn))
        self.coordinates_dphi_denergy = coords

    def _bin_nr_to_physical_coords(self,
                                   coordinates: Union[
                                       Sequence[np.ndarray],
                                       np.ndarray
                                   ],
                                   machine: 'Machine', recprof: int,
                                   deltaturn: int = 0) \
            -> Tuple[np.ndarray, np.ndarray]:
        """Function to convert from reconstructed phase space coordinates
        to physical units.

        Needed by the homogeneous_distribution routine.

        Parameters
        ----------
        coordinates: tuple (ndarray:float, ndarray:float)
            Coordinates of particles (I coordinate, J coordinate)
            provided as fractions of bins of the reconstructed phase space
            coordinate system.
        machine: Machine
            Machine object holding machine parameters and settings
        recprof: int
            Profile to reconstruct.
        deltaturn: int
            On offset in turn numbers to allow creation of distributions
            between two recorded profiles
        """
        turn = recprof * machine.dturns + deltaturn
        dphi = ((coordinates[0] + self.xorigin)
                * machine.h_num * machine.omega_rev0[turn] * machine.dtbin
                - machine.phi0[turn])
        denergy = (coordinates[1] - machine.synch_part_y) * self.dEbin
        return dphi, denergy

    # Assertions to assure that all needed fields are provided in the
    # given machine object
    def _assert_machine(self, machine: 'Machine'):
        needed_fieds = ['snpt', 'h_num', 'omega_rev0', 'eta0',
                        'dtbin', 'phi0', 'synch_part_y', 'dturns', 'phi12',
                        'nbins', 'beam_ref_frame', 'full_pp_flag',
                        'demax', 'e0', 'min_dt', 'max_dt', 'time_at_turn']
        asrt.assert_fields(
            machine, 'machine', needed_fieds, expt.MachineParameterError,
            'Did you remember to use machine.values_at_turns()?')


def filter_lost(xp: np.ndarray, yp: np.ndarray, img_width: int) \
        -> Tuple[np.ndarray, np.ndarray, int]:
    """Remove lost particles (particles that leaves the image width).

    Parameters
    ----------
    xp: ndarray
        2D array containing x-coordinates of all particles at each time frame.
        Particle coordinates must be given in phase space coordinates
        as integers. Shape: (nprofiles, nparts).
    yp: ndarray, int
        2D array containing y-coordinates of all particles at each time frame.
        Particle coordinates must be given in phase space coordinates
        as integers. Shape: (nprofiles, nparts).
    img_width: int
        Number of bins in measured profiles.

    Returns
    -------
    xp: ndarray
        2D array containing x-coordinates of all particles at each time frame.
        Now containing only particles inside of the image width.
        Shape: (nprofiles, nparts).
    yp: ndarray
        2D array containing y-coordinates of all particles at each time frame.
        Now containing only particles inside of the image width.
        Shape: (nprofiles, nparts).
    nr_lost_points: int
        Number of particles removed.

    """
    nr_lost_pts = 0
    
    # Find all particles outside of image width
    invalid_pts = conf.argwhere(conf.logical_or(xp >= img_width, xp < 0))

    if conf.size(invalid_pts) > 0:
        # Mark particle as invalid only once
        invalid_pts = conf.unique(invalid_pts[:, 1])
        # Save number of invalid particles
        nr_lost_pts = len(invalid_pts)

        valid_mask = ~conf.isin(conf.arange(xp.shape[1]), invalid_pts)

        xp = xp[:, valid_mask]
        yp = yp[:, valid_mask]

    if xp.size == 0:
        raise expt.InvalidParticleError(
            'All particles removed during filtering')

    return xp, yp, nr_lost_pts


def physical_to_coords(tracked_dphi: np.ndarray, tracked_denergy: np.ndarray,
                       machine: 'MachineABC', xorigin: float, dEbin: float) \
        -> Tuple[np.ndarray, np.ndarray]:
    """Function to convert from physical units ([rad], [eV]) to reconstructed
    phase space coordinates (bin numbers).

    The phase space coordinates are needed for the tomographic reconstruction
    routines.

    Parameters
    ----------
    tracked_dphi: ndarray
        2D array containing the phase difference [rad] relative to
        the synchronous particle for every particle at every time frame.
        Array shape: (nprofiles, nparticles)
    tracked_denergy: ndarray
        2D array containing the energy difference [eV] relative to
        the synchronous particle for every particle at every time frame.
        Array shape: (nprofiles, nparticles)
    machine: MachineABC
        machine object holding machine parameters, and settings for
        the reconstruction.
    xorigin: float
        The absolute difference (in bins) between phase=0 and the origin
        of the reconstructed phase space coordinate system.
    dEbin: float
        Energy size of bins in phase space coordinate system.

    Returns
    -------
    xp: ndarray
        2D array containing the x-coordinate for every particle
        at every time frame, given fractions of bins of
        the phase space coordinate system.
        Array shape: (nprofiles, nparticles)
    yp: ndarray
        2D array containing the y-coordinate for every particle
        at every time frame, given fractions of bins of
        the phase space coordinate system.
        Array shape: (nprofiles, nparticles)
    """
    if tracked_dphi.shape != tracked_denergy.shape:
        raise expt.InvalidParticleError(
            'Different shape of arrays containing phase and energies')
    nprof = tracked_denergy.shape[0]

    profiles = np.arange(nprof)

    turns = profiles * machine.dturns

    phi0_reshaped = conf.array(machine.phi0[turns].reshape(-1, 1))
    omega_rev0_reshaped = conf.array(machine.omega_rev0[turns].reshape(-1, 1))

    xp = ((tracked_dphi
           + phi0_reshaped)
          / (float(machine.h_num)
             * omega_rev0_reshaped
             * machine.dtbin) - xorigin)

    yp = (tracked_denergy
          / float(dEbin) + machine.synch_part_y)

    return xp, yp


def ready_for_tomography(xp: np.ndarray, yp: np.ndarray, nbins: int) \
        -> Tuple[np.ndarray, np.ndarray]:
    """Function to prepare tracked particles tomography routine.

    Handy if particles are tracked using the functions of the
    :mod:`longitudinal_tomography.tracking.tracking` module.

    **Function does the following actions:**

    * Removes particles leaving the image width.
    * Transposes the coordinate arrays from (nprofiles, nparts)\
    to (nparts, nprofiles).
    * Casts coordinates to integers.

    The returned coordinates are ready to be used in the tomographic
    reconstruction routines.

    Parameters
    ----------
    xp: ndarray
        2D array containing the x-coordinate for every particle
        at every time frame, given fractions of bins of
        the phase space coordinate system.
        Array shape: **(N, M)**, where N is the number of profiles,
        and M is the number of particles.
    yp: ndarray, float
        2D array containing the y-coordinate for every particle
        at every time frame, given fractions of bins of
        the phase space coordinate system.
        Array shape: **(N, M)**, where N is the number of profiles,
        and M is the number of particles.

    Returns
    -------
    xp: ndarray
        2D array containing the x-coordinate for every particle
        at every time frame, given in phase space coordinates.
        Array shape: **(M, N)**, where N is the number of profiles,
        and M is the number of particles.
    yp: ndarray
        2D array containing the y-coordinate for every particle
        at every time frame, given in phase space coordinates.
        Array shape: **(M, N)**, where N is the number of profiles,
        and M is the number of particles.
    """
    xp, yp, lost = filter_lost(xp, yp, nbins)
    log.info(f'number of lost particles: {lost}')

    xp = xp.astype(conf.int32).T
    yp = yp.astype(conf.int32).T

    return xp, yp


# Function to check that coordinates are valid
def _assert_coordinates(coordinates: Sequence[np.ndarray]) \
        -> Tuple[np.ndarray, np.ndarray]:
    if not hasattr(coordinates, '__getitem__'):
        raise expt.InvalidParticleError('coordinates should be a sequence')
    if not len(coordinates) == 2:
        raise expt.InvalidParticleError('Two arrays of coordinates should be'
                                        'provided')
    for coord in coordinates:
        if not hasattr(coord, '__iter__'):
            raise expt.InvalidParticleError('coordinates should be iterable')
    if not len(coordinates[0]) == len(coordinates[1]):
        raise expt.InvalidParticleError(
            'arrays holding coordinates of x and y axis '
            'should have the same length')
    return coordinates[0], coordinates[1]



import os

import numpy as np
import longitudinal_tomography.tracking.particles as parts
import longitudinal_tomography.tracking.tracking as tracking
import longitudinal_tomography.utils.tomo_input as tomoin
import longitudinal_tomography.utils.tomo_output as tomoout
import longitudinal_tomography.data.data_treatment as dtreat


def discrepancy(nbins, nprofs, dwaterfall):
    return np.sqrt(np.sum(dwaterfall ** 2) / (nbins * nprofs))


# -----------------------------------------------------------------------------
# Data loading or generation, not part of the example. Skip ahead.
# -----------------------------------------------------------------------------
this_dir = os.path.realpath(os.path.dirname(__file__))
resource_dir = os.path.join(this_dir, 'resources')

xp_file = os.path.join(resource_dir, 'INDIVShavingC325_xcoords.npy')
yp_file = os.path.join(resource_dir, 'INDIVShavingC325_ycoords.npy')
waterfall_file = os.path.join(resource_dir, 'INDIVShavingC325_waterfall.npy')

if os.path.exists(xp_file) and os.path.isfile(xp_file):
    xp = np.load(xp_file)
    yp = np.load(yp_file)
    waterfall = np.load(waterfall_file)
else:
    # construct pre-tracked data, this routine is not part of the example
    # the saved numpy arrays take up too much space in the repo, instead
    # they are generated when needed
    ex_dir = os.path.split(this_dir)[0]
    in_file_pth = os.path.join(ex_dir, 'input_files', 'INDIVShavingC325.dat')

    with open(in_file_pth, 'r') as file:
        raw_params, raw_data = tomoin._split_input(file.readlines())

    machine, frames = tomoin.txt_input_to_machine(raw_params)
    machine.values_at_turns()
    measured_waterfall = frames.to_waterfall(raw_data)

    profiles = tomoin.raw_data_to_profiles(
        measured_waterfall, machine, frames.rebin, frames.sampling_time)
    profiles.calc_profilecharge()

    if profiles.machine.synch_part_x < 0:
        fit_info = dtreat.fit_synch_part_x(profiles)
        machine.load_fitted_synch_part_x_ftn(fit_info)

    reconstr_idx = machine.filmstart

    # Tracking...
    tracker = tracking.Tracking(machine)

    xp, yp = tracker.track(reconstr_idx)

    # Converting from physical coordinates ([rad], [eV])
    # to phase space coordinates.
    if not tracker.self_field_flag:
        xp, yp = parts.physical_to_coords(
            xp, yp, machine, tracker.particles.xorigin,
            tracker.particles.dEbin)

    # Filters out lost particles, transposes particle matrix,
    # casts to np.int32.
    xp, yp = parts.ready_for_tomography(xp, yp, machine.nbins)

    waterfall = profiles.waterfall

    os.makedirs(resource_dir, exist_ok=True)
    np.save(xp_file, xp)
    np.save(yp_file, yp)
    np.save(waterfall_file, waterfall)

# -----------------------------------------------------------------------------
# Begin main example
# -----------------------------------------------------------------------------

niterations = 20
nprofs = waterfall.shape[0]
nbins = waterfall.shape[1]
nparts = xp.shape[0]
rec_tframe = 0

# Remove comment to track using longitudinal_tomography routine:
# ------------------------------------------------------
# import longitudinal_tomography.tomography.tomography_cpp as tomography
# import sys
# longitudinal_tomography = tomography.TomographyCpp(waterfall, xp)
# weight = longitudinal_tomography.run(niter=niterations)
# image = tomoout.create_phase_space_image(
#             xp, yp, weight, nbins, rec_tframe)
# tomoout.show(image, longitudinal_tomography.diff, waterfall[rec_tframe])
# sys.exit()
# ------------------------------------------------------

# Waterfall must be normalized, and negatives
# reduced to zeroes before reconstruction.
flat_profs = waterfall.copy()
flat_profs = flat_profs.clip(0.0)

flat_profs /= np.sum(flat_profs, axis=1)[:, None]
flat_profs = np.ascontiguousarray(flat_profs.flatten()).astype(np.float64)

# Normalizing waterfall in profiles used for comparing
waterfall /= np.sum(waterfall, axis=1)[:, None]

# In order to use the cpp reconstruction routines, the profile
# arrays must be flattened. x-coordinates must be adjusted for this.
flat_points = xp.copy()
for i in range(nprofs):
    flat_points[:, i] += nbins * i

# Reconstructing phase space
weight = np.zeros(nparts)
rec_wf = np.zeros(waterfall.shape)

# Initial estimation of weight factors using (flattened) measured profiles.
weight = libtomo.back_project(weight, flat_points, flat_profs,
                              nparts, nprofs)
weight = weight.clip(0.0)

diff = []
for i in range(niterations):
    # Projection from phase space to time projections
    rec_wf = libtomo.project(rec_wf, flat_points, weight, nparts,
                             nprofs, nbins)

    # Normalizing reconstructed waterfall
    rec_wf /= np.sum(rec_wf, axis=1)[:, None]

    # Finding difference between measured and reconstructed waterfall
    dwaterfall = waterfall - rec_wf

    # Setting to zero for next round
    rec_wf[:] = 0.0

    # Calculating discrepancy
    diff.append(np.sqrt(np.sum(dwaterfall ** 2) / (nbins * nprofs)))

    # Back projecting using the difference between measured and rec. waterfall
    weight = libtomo.back_project(weight, flat_points, dwaterfall.flatten(),
                                  nparts, nprofs)
    weight = weight.clip(0.0)

    print(f'Iteration: {i:3d}, discrepancy: {diff[-1]:3E}')

# Finding last discrepancy...
rec_wf = libtomo.project(rec_wf, flat_points, weight, nparts, nprofs, nbins)
rec_wf /= np.sum(rec_wf, axis=1)[:, None]
dwaterfall = waterfall - rec_wf
diff.append(np.sqrt(np.sum(dwaterfall ** 2) / (nbins * nprofs)))
print(f'Iteration: {i + 1:3d}, discrepancy: {diff[-1]:3E}')

# Creating image for fortran style presentation of phase space
image = tomoout.create_phase_space_image(
    xp, yp, weight, nbins, rec_tframe)
tomoout.show(image, diff, waterfall[rec_tframe])
